"
NAME		RemoteInvocation.st
AUTHOR		Wang Xu
CONTRIBUTOR	Wang Xu <wang@dcs.qmw.ac.uk>
FUNCTION	Support remote invocations between Smalltalk images.
ST-VERSIONS	R4.0
PREREQUISITES	
CONFLICTS	
DISTRIBUTION	world
VERSIONS	R	1.5
DATE		1 Nov. 1992
SUMMARY	

		Remote Invocation for Distributed Smalltalk

RemoteInvocation.st includes several classes that support the transparent
remote invocation between running Smalltalk-80 images. It is the kernal 
part of a version of Distributed and Persistent Smalltalk developed and used
in Distributed System Research Group, Queen Mary & Westfield College, which
is supported by Spirit Project.
	
This version is implemented purely in Smalltalk source (e.g. no virtual 
machine changes and C agents). To installation, just file in it and read
the document messages in the class RemoteInvocationServer.

There is another package, RemoteTools.st, which includes implementations 
of the remote browse and the remote inspector. The remote browser allows
you to browse system organizations of other images. It alos allows directly
fileIn the sourece from other images.  The remote inspector support 
inspecting remote objects. Both remote browser and remote inspector
use transparent remote invocations. They are good demonstrations for using
remote invocation.

Any comments and bug report/fix are welcome and should be directed to me 
but I may not be able to reply all the messages came in.

We also have several technical reports about our projects which are available
on request.


Wang Xu					Email: wang@dcs.qmw.ac.uk
Department of Computer Science		Tel: +44 071 975 5244
Queen Mary & Westfield College		Fax: +44 081 980 6533
Mile End Road, London E1 4NS
UK
"!

'From Objectworks(r)\Smalltalk, Release 4 of 25 October 1990 on 3 November 1992 at 10:35:57 am'!



| s | s := '
		Remote Invocation Version 1.5
		For ObjectWorks/Smalltalk-80R4

		Wang Xu.
		Queen Mary and Westfield College.
		University of London.
		14 June 1992.
'.
	Transcript show: s.!


	(DialogView confirm:
'It is recommended that the previous version of this
package be removed.  Doing so involves removing all
methods in category ''remote invocation packing unpacking'' 
and removing classes in Remote-Invocation and all their subclasses.
Do you wish to do so now?') ifTrue: 
			[#('remote invocation' 'packing' 'unpacking') do:
					[:category | category asSymbol.
					Smalltalk allBehaviorsDo: 
						[:class | (class organization listAtCategoryNamed: category) do:
							[:sel | class removeSelector: sel]]].
			(Smalltalk at: #Proxy ifAbsent:[]) isNil
				ifFalse:[Proxy releaseProxyTable].
			#('Remote-Invocation' ) 
				do: [:cat | (SystemOrganization superclassOrder:  cat )do: 
						[:class | Smalltalk removeClassNamed: class name]]]!

Object subclass: #TestObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Remote-Invocation'!


!TestObject methodsFor: 'accessing'!

testMessage1: para
	^para!

testMessage2: para
	^para basicAt: 1!

testMessage3: para
	| t |
	t := para basicAt: 2.
	Transcript cr; show: 'Test 3 returns ', t printString.
	^t!

testMessage4: para
	^#white!

testMessage5: para
	^'Test return message for testMessage5'! !

nil subclass: #Encapsulator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Remote-Invocation'!
Encapsulator comment:
'I am an abstract superclass for encapsulators. 
Encapsulators are objects that don''t understand any messages 
other than doesNotUnderstand:. They typically forward the 
message to some other object performing some other computation 
in the process.'!


!Encapsulator methodsFor: 'primitives'!

class
	"Answer the object which is the receiver's class."

	<primitive: 111>
	self primitiveFailed! !

!Encapsulator methodsFor: 'shared access'!

doesNotUnderstand: aMessage
	"I am an abstract class; inform the user that my subclasses 
should implement doesNotUnderstand:"

	^Object messageNotUnderstoodSignal
		raiseRequestWith: aMessage
		errorString: 'Message not understood: ' , aMessage selector! !

!Encapsulator methodsFor: 'primitive identification'!

isReallyEncapsulator
	"This message is useful and can be overridden by subclasses if not wanted"
	^true! !

!Encapsulator methodsFor: 'finalization'!

executor
	"Answer an object to execute the finalization for the receiver.
	Subclasses might want to override this method."

	^self!

finalize
	"Finalize the object. This message is sent by WeakDictionary.
	The default is to do nothing. Subclasses might want to override 
	this method."

	^self! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Encapsulator class
	instanceVariableNames: ''!


!Encapsulator class methodsFor: 'class initialization'!

initialize
	"Ensure my instances inherit no messages by setting my superclass to nil.
	 Inform Smalltalk that I am now a root of a class heirarchy"

	"Encapsulator initialize"

	(self == Encapsulator and: [superclass ~~ nil])
		ifTrue: [
			superclass removeSubclass: self.
			superclass _ nil]! !

!Encapsulator class methodsFor: 'fileIn/Out'!

fileOutOn: aFileStream moveSource: moveSource toFile: fileIndex
	"I have to be clever when filiong out, because I don't have a superclass.
	 So that I may be filed in again, when being filed out I pretend my super is Object"

	| startPosition endPosition |
	(superclass == nil and: [moveSource not])
		ifTrue: [
			startPosition _ aFileStream position.
			aFileStream nextPutAll: '   '		
"String new: 'Object' size - 'nil' size withAll: $  "].

	super fileOutOn: aFileStream moveSource: moveSource toFile: fileIndex.

	(superclass == nil and: [moveSource not])
		ifTrue: [
			endPosition _ aFileStream position.
			aFileStream
				position: startPosition;
				nextPutAll: 'Object';
				position: endPosition]! !

Encapsulator subclass: #Proxy
	instanceVariableNames: 'globalId accessPoint timeout '
	classVariableNames: 'DistributedGarbageCollectionOn ProxyTable '
	poolDictionaries: ''
	category: 'Remote-Invocation'!
Proxy comment:
'Instances of this class represent local agents of remote objects.

To a message sender, a proxy maintains the remote object message 
protocol by forwarding the message to the remote object and blocks 
the calling thread until the invocation returns. 
The message forwarding is done by sending the remote object gid 
and the packed message to the remote side invocation server 
(addressed by accessPoint of the proxy). (See doesNotUnderstand)

Proxies also has a set of meta methods which are interpreted locally.

Intance variables:
	globalId			<GlobalId>			Remote object''s unique object identifier
	communicator 	<Communicator>	The communiction protocol implementor
	accessPointer	<IPSocketAddress>	The remote object''s access point
	timeout			<SmallInteger>		Timeout value in seconds

class variables:
	ProxyTable		<WeakDic>	A map between gids and proxies, 
which is used for make sure the proxy for a remote object is unique in local 
image and be garbage collected when all the local references (except the ProxyTable) 
to it are gone. 

Notes:
Proxies should be created automaticly and implicitly by remote invocation 
server or other distributing servers.
'!


!Proxy methodsFor: 'initialize - finaliza'!

finalize
	"Finalize the proxy"
	
	self proxyFinalize!

proxyFinalize
	"Finalize the proxy"

	
	"For distributed garbage collection:"
	ProxyTable removeKey: globalId.
	RIServer sendDecRefOf: globalId at: accessPoint!

proxyInitialize: gid accessPoint: socketAddress
	"Initialize the proxy"

	DistributedGarbageCollectionOn ifTrue: ["Distrubuted Garbage Collection is On"
		(RIServer
			sendIncRefOf: gid
			at: socketAddress
			timeout: 3) key ~= 'SUCCESS' 
				ifTrue: ["Increase the remote reference count has failed" ^nil]].

	globalId := gid.
	accessPoint := socketAddress.	"Set the default timeout value in seconds"
	timeout := 60.	"Add my self in the ProxyTable"
	ProxyTable at: gid put: self.
	^self! !

!Proxy methodsFor: 'remote copy'!

remoteCopy
	"Make a local copy of the remote object"
	"Not supported in this simplified version"
"
	| reply |

	reply := RIServer
				remoteCopy: globalId
				at: accessPoint
				timeout: timeout.

	reply == nil | (reply class ~~ Association)
		ifTrue: [
			Transcript cr; show: 'Error: remote invocation return nil.'.
			^nil].

	reply key = 'SUCCESS' ifTrue: [^reply value].

	reply key = 'ERROR'
		ifTrue: [
			Transcript cr; show: 'Remote invocation error: ' , reply value printString.
			Transcript cr; show: 'Gid: ' , globalId printString.
			^nil]"! !

!Proxy methodsFor: 'message forwarding'!

doesNotUnderstand: aMessage 
	"Forward the message and return the reply"

	| reply |
	
	reply := RIServer 
			forwardAndGetReply: aMessage 
			to: globalId 
			at: accessPoint 
			timeout: timeout.
	reply == nil | (reply class ~~ Association)
		ifTrue:[Transcript cr; show: 'Error: remote invocation return nil.'.^nil].

	reply key = 'SUCCESS'
		ifTrue:[^reply value].
	
	reply key = 'ERROR' 
		ifTrue: [
			Transcript cr; show: 'Remote invocation reply error: ' , reply value printString.
			Transcript cr; show: 'Gid: ' , globalId printString.
			^self error: reply value].

	reply key = 'APCHANGED'
		ifTrue:[
			self PObecome: reply value.
			^self perform: aMessage selector withArguments: aMessage arguments].
			"We need self resend: aMessage"!

remoteClass
	"Forwarding  #class.
	#class is a hard binding message and it will not intercepted by
	aProxy "

	^self doesNotUnderstand: (Message selector: #class)! !

!Proxy methodsFor: 'meta methods'!

= anObject
	"Proxies are equal if their gids are equal"

	^anObject = globalId!

hash
	"Answer a SmallInteger whose value reflects the proxy virtual identity."

	^globalId hash!

identityHash
	"Answer a SmallInteger whose value reflects the identity, not the contents,
	of the receiver.  Essential.  See documentation in Object metaclass."

	<primitive: 75>
	self primitiveFailed!

inspect
	"Remote inspection: inspect the remote object"

	| class foo |

	class := Smalltalk at: (self remoteClass name).
	class isNil ifTrue: [^InspectorView openRemote: (RemoteInspector inspect: self)].
	foo := class new.
	(foo isKindOf: Dictionary)
		ifTrue: [^InspectorView openRemote: (RemoteDictionaryInspector inspect: self)].
	(foo isKindOf: OrderedCollection)
		ifTrue: [^InspectorView openRemote: (RemoteOrderedCollectionInspector inspect: self)].
	(foo isKindOf: SequenceableCollection)
		ifTrue: [^InspectorView openRemote: (RemoteSequenceableCollectionInspector inspect: self)].
	InspectorView openRemote: (RemoteInspector inspect: self)!

isReallyProxy
	^true!

proxyAccessPoint
	^accessPoint!

proxyEncapsulatedObject
	^globalId!

proxyHostName
	^accessPoint hostName!

proxySetTimeout: aSmallInteger
	"Set the timeout to aSmallinteger seconds"

	timeout := aSmallInteger.!

proxyTimeoutValue
	^timeout! !

!Proxy methodsFor: 'remote invocation'!

isSentByValue
	"When the receiver is a argument in a remote message 
	 answer whether they should be sent by value or by proxy. 
	 The receiver should be well known and immutable 
	 so should be sent by value."


	^true!

putValueOn: stream
	stream nextGlobalIdPut: globalId.
	stream nextIPSocketAddressPut: accessPoint.! !

!Proxy methodsFor: 'error handling'!

halt
	"This is a simple message to use for inserting breakpoints during debugging.
	The debugger is opened by raising a signal. This gives a chance to restore
	invariants related to multiple processes."

	Object haltSignal 
				raiseRequestWith: thisContext
				errorString: 'Halt encountered.'

	"nil halt."!

halt: aString 
	"This message can be used for inserting breakpoints during debugging.
	It creates and schedules a Debugger with the argument, aString, as 
	the label. The debugger is opened by raising a signal. This gives a chance
	to restore invariants related to multiple processes."

	Object haltSignal 
				raiseRequestWith: thisContext
				errorString: aString

	"nil halt: 'Test of halt:.'."! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Proxy class
	instanceVariableNames: ''!


!Proxy class methodsFor: 'class initialization'!

initialize
	"Proxy initialize"

	"If you want the distributed garbage collection, trun it on by:
	DistributedGarbageCollectionOn := true"

	DistributedGarbageCollectionOn := false.! !

!Proxy class methodsFor: 'instance creation'!

getValueFrom: stream
	| gid ap   |
	gid := stream nextGlobalId.
	ap := stream nextIPSocketAddress.
	^RIServer exportObjByGid: gid 
				ifAbsent: [self newProxy: gid accessPoint: ap]!

newProxy: gid accessPoint: ipSocketAddress
	"Check if the proxy already created then renturn the old proxy
	otherwise return a new proxy. 
	Proxy table is needed in the distributed garbage collection."

	| proxy | 
	^ProxyTable at: gid ifAbsent:[
		proxy := (super new) proxyInitialize: gid accessPoint: ipSocketAddress.
		proxy]! !

!Proxy class methodsFor: 'proxy table access'!

initializeProxyTable
	DistributedGarbageCollectionOn
		ifTrue: ["Distriubted Garbage Collection is On"
			^ProxyTable := WeakDic new: 1024]
		ifFalse: ["Distributed Garbage Collection is Off"
			^ProxyTable := Dictionary new: 1024]!

releaseProxyTable
	DistributedGarbageCollectionOn ifTrue: ["For distributed garbage collection"
		ProxyTable isNil ifFalse: [ProxyTable do: 
				[:proxy | 
				proxy proxyFinalize]]].
	ProxyTable := nil! !

!Proxy class methodsFor: 'meta accessing'!

instancesSentByValue
	^true! !

ByteArray variableByteSubclass: #RemoteMessage
	instanceVariableNames: ''
	classVariableNames: 'AddressSize BufferSize GidIndex GidSize HeaderSize 
MaxSize MessageIndex MessageSizeIndex ReplyIdIndex SenderAddressIndex TagIndex '
	poolDictionaries: ''
	category: 'Remote-Invocation'!
RemoteMessage comment:
'Instances of this class are remote messages which are ready to 
send through the UDP socket.
Each remote messages has a header and a body.
Header include:
	Receiver''s gid;
	Message tag (invocation/reply);
	Message body size;
	Sender''s waiting semaphore id;
	Sender''s socket address;

Body for invocation message include:
	Message selector;
	Parameter objects.

Body for reply messae include:
	Reply state code;
	Reply object.

Body for error message include:
	Reply state code; (ERROR)
	A error string.
	Should be a exception which can be raised at destination.

Class Variables:
	MaxSize		<SmallInteger>			"Maxmun single message size: 4056"
	GidIndex		<SmallInteger>			"Object global id at 1-8"
	TagIndex		<SmallInteger>			"Message tag: invo 0, reply 1, error 2"
	MessageSizeWordIndex		<SmallInteger> 	"Real packed message length at: 10-11."
	SenderSemaphoreIndex		<SmallInteger> 	"Sender Semaphore at: 12-15"
	SenderAddressIndex		<SmallInteger> 	"Sender address at: 16 - 31"
	HeaderSize	<SmallInteger>			"32 bytes"
	MessageIndex	<SmallInteger>		"Where the packed message starts"'!


!RemoteMessage methodsFor: 'accessing'!

body
	^self copyFrom: MessageIndex to: self getSize!

header
	^self copyFrom: 1 to: HeaderSize!

headerSize
	^HeaderSize!

messageStart
	^MessageIndex! !

!RemoteMessage methodsFor: 'header accessing'!

getGid
	^(GlobalId basicNew: 8)
		replaceFrom: 1
		to: 8
		with: self
		startingAt:  GidIndex 

"(RemoteMessage newAsBuffer) setGid: GlobalId newId; getGid."!

getReplyId
	
	^self doubleWordAt: ReplyIdIndex!

getSenderAddress
	| fa  i  size | 
	i := 0.
	fa := IPSocketAddress new.
	size := fa basicSize.
	[(i := i + 1) <= size]
		whileTrue: [ fa byteAt: i  put: (self at: (i + SenderAddressIndex - 1))].
	^fa
"RemoteMessage newAsBuffer setSenderAddress: (IPSocketAddress hostName: 'sake'); getSenderAddress"!

getSize
	^self doubleWordAt: MessageSizeIndex!

getTag
	^self at: TagIndex!

setGid: gid
	^self 
		replaceFrom: GidIndex 
		to: GidIndex + 7 
		with: gid 
		startingAt: 1!

setReplyId: id
	
	^self doubleWordAt: ReplyIdIndex put: id!

setSenderAddress: sa 
	| i size|
	i := 0.
	size := sa basicSize.
	[(i := i + 1) <= size]
		whileTrue: [ self at: (i + SenderAddressIndex - 1)put: ( sa byteAt: i)]!

setSize: size
	self doubleWordAt: MessageSizeIndex put: size!

setTag: t
	^self at: TagIndex put: t! !

!RemoteMessage methodsFor: 'mashalling'!

doubleAt: byteIndex
	"Answer a double-precision (64 bit) floating point quantity
	starting at the given byteIndex."

	<primitive: 550>
	"The only way we can do this is by
	constructing the Double byte-by-byte."
	| f |
	f := 0 asDouble.  "make a Double"
	1 to: f basicSize do:
		[ :i |	f basicAt: i put: (self byteAt: byteIndex+i-1)].
	^f!

doubleAt: byteIndex put: aDouble
	"Store a double-precision (64 bit) floating point quantity
	starting at the given byteIndex."

	<primitive: 551>
	"The only way we can do this is by
	decomposing the long float byte-by-byte."
	| f |
	f := aDouble asDouble.
	1 to: f basicSize do:
		[ :i |	self byteAt: byteIndex+i-1 put: (f basicAt: i)].
	^f!

floatAt: byteIndex
	"Answer a single-precision (32 bit)
	floating point quantity
	starting at the given byteIndex."

	<primitive: 548>
	"The only way we can do this is by
	constructing the Float byte-by-byte."
	| f |
	f := 0.0 + 0.0.  "make a Float"
	f basicAt: 1 put: (self byteAt: byteIndex).
	f basicAt: 2 put: (self byteAt: byteIndex+1).
	f basicAt: 3 put: (self byteAt: byteIndex+2).
	f basicAt: 4 put: (self byteAt: byteIndex+3).
	^f!

floatAt: byteIndex put: aFloat 
	"Store a single-precision (32 bit) 
	floating point quantity 
	starting at the given byteIndex."

	<primitive: 549>
	"The only way we can do this is by 
	storing the Float byte-by-byte."
	| float |
	float := aFloat asFloat.
	self byteAt: byteIndex put: (float basicAt: 1).
	self byteAt: byteIndex + 1 put: (float basicAt: 2).
	self byteAt: byteIndex + 2 put: (float basicAt: 3).
	self byteAt: byteIndex + 3 put: (float basicAt: 4).
	^float!

longAt: byteIndex
	"Answer a signed long (32 bit) quantity
	starting at the given byteIndex."

	<primitive: 546>
	^self longAt: byteIndex bigEndian: self class isBigEndian!

longAt: byteIndex bigEndian: isBE
	"Answer a signed long (32 bit) quantity starting at the given byteIndex."

	| b0 b1 b2 w h |
	isBE
		ifTrue:
			[b0 := self byteAt: byteIndex.  
			b1 := self byteAt: byteIndex+1.  
			b2 := self byteAt: byteIndex+2.  
			w := self byteAt: byteIndex+3]
		ifFalse:
			[b0 := self byteAt: byteIndex+3.  
			b1 := self byteAt: byteIndex+2.  
			b2 := self byteAt: byteIndex+1.  
			w := self byteAt: byteIndex].
	"Following sequence minimizes LargeInteger arithmetic for small results."
	h := ((b0 bitAnd: 16r7F) - (b0 bitAnd: 16r80) bitShift: 8) + b1.
	b2=0 ifFalse: [w := (b2 bitShift: 8) + w].
	h=0 ifFalse: [w := (h bitShift: 16) + w].
	^w!

longAt: byteIndex put: aLong
	"Store a signed long (32 bit) quantity
	starting at the given byteIndex."

	<primitive: 547>
	^self longAt: byteIndex put: aLong bigEndian: self class isBigEndian!

longAt: byteIndex put: aLong bigEndian: isBE
	"Store a signed long (32 bit) quantity starting at the given byteIndex."

	| b0 b1 b2 b3 |
	b0 := aLong bitShift: -24.
	  b0 := (b0 bitAnd: 16r7F) - (b0 bitAnd: -16r80).
	b1 := (aLong bitShift: -16) bitAnd: 16rFF.
	b2 := (aLong bitShift: -8) bitAnd: 16rFF.
	b3 := aLong bitAnd: 16rFF.
	isBE
		ifTrue:
			["Only the first store can fail."
			self byteAt: byteIndex put: b0.  
			self byteAt: byteIndex+1 put: b1.
			self byteAt: byteIndex+2 put: b2.  
			self byteAt: byteIndex+3 put: b3]
		ifFalse:
			["Only the first store can fail."
			self byteAt: byteIndex+3 put: b0.  
			self byteAt: byteIndex+2 put: b1.
			self byteAt: byteIndex+1 put: b2.  
			self byteAt: byteIndex put: b3].
	^aLong!

shortAt: byteIndex
	"Answer a signed short (16 bit) quantity
	starting at the given byteIndex."

	<primitive: 542>
	| word |
	word := self unsignedShortAt: byteIndex.
	^(word bitAnd: 16r7FFF) - (word bitAnd: 16r8000)!

shortAt: byteIndex put: aShort
	"Store a signed short (16 bit) quantity
	starting at the given byteIndex."

	<primitive: 543>
	self unsignedShortAt: byteIndex put:
		(aShort bitAnd: 16r7FFF) - (aShort bitAnd: -16r8000).
	^aShort!

shortAt: byteIndex put: aShort bigEndian: isBE
	"Store a signed short (16 bit) quantity starting at the given byteIndex."

	self unsignedShortAt: byteIndex put:
		(aShort bitAnd: 16r7FFF) - (aShort bitAnd: -16r8000) bigEndian: isBE.
	^aShort!

stringAt: byteIndex 
	"Answer a new String at the byteIndex. "

	| aString length index |
	index := byteIndex - 1.
	length := self shortAt: (index := index + 1).	
	aString := String new: length.
	1 to: length do: [:i | aString at: i put: (self basicAt: (index := index + 1))].
	^aString!

stringAt: byteIndex put: aString
	"Answer a new String at the byteIndex. "

	| index size |
	self shortAt: byteIndex put: (size := aString size).
	self replaceBytesFrom: (index := byteIndex + 2) to: index  + size with: aString startingAt: 1.
	^index + size!

unsignedLongAt: byteIndex
	"Answer an unsigned long (32 bit) quantity
	starting at the given byteIndex."

	<primitive: 544>
	^self unsignedLongAt: byteIndex bigEndian: self class isBigEndian!

unsignedLongAt: byteIndex bigEndian: isBE
	"Answer an unsigned long (32 bit) quantity starting at the given byteIndex."

	| b0 b1 b2 w |
	isBE
		ifTrue:
			[b0 := self byteAt: byteIndex.  
			b1 := self byteAt: byteIndex+1.  
			b2 := self byteAt: byteIndex+2.  
			w := self byteAt: byteIndex+3]
		ifFalse:
			[b0 := self byteAt: byteIndex+3.  
			b1 := self byteAt: byteIndex+2.  
			b2 := self byteAt: byteIndex+1.  
			w := self byteAt: byteIndex].
	"Following sequence minimizes LargeInteger arithmetic for small results."
	b2=0 ifFalse: [w := (b2 bitShift: 8) + w].
	b1=0 ifFalse: [w := (b1 bitShift: 16) + w].
	b0=0 ifFalse: [w := (b0 bitShift: 24) + w].
	^w!

unsignedLongAt: byteIndex put: anUnsignedLong
	"Store an unsigned long (32 bit) quantity
	starting at the given byteIndex."

	<primitive: 545>
	^self unsignedLongAt: byteIndex put: anUnsignedLong bigEndian: self class isBigEndian!

unsignedLongAt: byteIndex put: anUnsignedLong bigEndian: isBE
	"Store an unsigned long (32 bit) quantity starting at the given byteIndex."

	| b0 b1 b2 b3 |
	b0 := anUnsignedLong bitShift: -24.
	b1 := (anUnsignedLong bitShift: -16) bitAnd: 16rFF.
	b2 := (anUnsignedLong bitShift: -8) bitAnd: 16rFF.
	b3 := anUnsignedLong bitAnd: 16rFF.
	isBE
		ifTrue:
			["Only the first store can fail."
			self byteAt: byteIndex put: b0.  
			self byteAt: byteIndex+1 put: b1.
			self byteAt: byteIndex+2 put: b2.  
			self byteAt: byteIndex+3 put: b3]
		ifFalse:
			["Only the first store can fail."
			self byteAt: byteIndex+3 put: b0.  
			self byteAt: byteIndex+2 put: b1.
			self byteAt: byteIndex+1 put: b2.  
			self byteAt: byteIndex put: b3].
	^anUnsignedLong!

unsignedShortAt: byteIndex
	"Answer an unsigned short (16 bit) quantity
	starting at the given byteIndex."

	<primitive: 540>
	^self unsignedShortAt: byteIndex bigEndian: self class isBigEndian!

unsignedShortAt: byteIndex put: anUnsignedShort
	"Store an unsigned short (16 bit) quantity
	starting at the given byteIndex."

	<primitive: 541>
	^self unsignedShortAt: byteIndex put: anUnsignedShort bigEndian: self class isBigEndian! !

!RemoteMessage methodsFor: 'comparing'!

= anObject
	"Answer whether the bytes of the receiver and argument are the same."

	(anObject class isBits and: [self sizeInBytes = anObject basicSize])
		ifFalse: [^false].
	1 to: self sizeInBytes do: [:i | (self byteAt: i)
			= (anObject basicAt: i) ifFalse: [^false]].
	^true!

hash
	"Answer a SmallInteger unique to the receiver."

	self sizeInBytes = 0 ifTrue: [^17171].
	^(self byteAt: 1) hash + (self byteAt: self sizeInBytes) hash + self sizeInBytes hash! !

!RemoteMessage methodsFor: 'byte order reversal'!

reverseByteOrder
	"Only subclasses know what their byte interpretation is.
	Default is no interpretation."
	^self!

swapLongAt: index 
	"Used in migrating the receiver onto a new host platform"

	| tmp |
	tmp := self byteAt: index.
	self byteAt: index put: (self byteAt: index + 3).
	self byteAt: index + 3 put: tmp.
	tmp := self byteAt: index + 1.
	self byteAt: index + 1 put: (self byteAt: index + 2).
	self byteAt: index + 2 put: tmp!

swapShortAt: index 
	"Used in migrating the receiver onto a new host platform"

	| tmp |
	tmp := self byteAt: index.
	self byteAt: index put: (self byteAt: index + 1).
	self byteAt: index + 1 put: tmp! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

RemoteMessage class
	instanceVariableNames: ''!


!RemoteMessage class methodsFor: 'classs intialization'!

initialize
	"RemoteMessage initialize"

	BufferSize := 4100.				"Maxmun single message size"
	GidIndex := 1.					"Object global id at 1-8"
	GidSize	:= 8.
	TagIndex := 9.					"Message tag"
	MessageSizeIndex := 10.		"Real packed message length at: 10-13."
	ReplyIdIndex := 14.				"ReplyId at: 14-17"
	SenderAddressIndex := 18.		"Sender address at: 18 - 33"
	AddressSize	:= 16.
	HeaderSize := 64.					
	MessageIndex := 65.				"Where the packed message starts"! !

!RemoteMessage class methodsFor: 'instance creation'!

newAsBuffer
	^self new: BufferSize!

newheader
	^self new: HeaderSize! !

IdentityDictionary variableSubclass: #ObjectIdTable
	instanceVariableNames: 'indexDictionary referenceArray '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Remote-Invocation'!
ObjectIdTable comment:
'I am a specialization of IdentityDictionary whose values are a set 
(i.e. I am a one-to-one mapping, not a one-to-many mapping & hence have 
a single inverse for each value).  I maintain a reverse map from value 
to key index so that keyAtValue: is as efficient as at:.

Instance Variables:
	indexDictionary	<IdentityDictionary of <Object> -> <Integer>>
	:- map from each value to their respective key''s index'!


!ObjectIdTable methodsFor: 'accessing'!

at: key persistentPut: value
	"Set the value at key to be value.
	 Record value -> index in the indexDictionary
	Persistent item will not be reclaimed by 
	distributed garbage collection."
 
	| index |
	index := self findKeyOrNil: key.
	(self basicAt: index) == nil ifTrue: [
		tally := tally + 1.
		self basicAt: index put: key].
	valueArray at: index put: value.
	indexDictionary at: value put: index.
	self fullCheck.
	referenceArray at: index put: 1.
	^value!

at: key put: value
	"Set the value at key to be value.
	 Record value -> index in the indexDictionary"
 
	| index |
	index := self findKeyOrNil: key.
	(self basicAt: index) == nil ifTrue: [
		tally := tally + 1.
		self basicAt: index put: key].
	valueArray at: index put: value.
	indexDictionary at: value put: index.
	self fullCheck.
	^value!

keyAtValue: value ifAbsent: exceptionBlock
	"Answer the key whose value equals the argument, value.  If there is none, 
	answer the result of evaluating exceptionBlock.  
	This is implemented efficiently because we maintain a dictionary of value -> index"

	| index theKey |
	index := indexDictionary at: value ifAbsent: [^exceptionBlock value].
	value == (valueArray at: index) ifFalse: [self invariantViolated].
	(theKey := self basicAt: index) == nil ifTrue: [self invariantViolated].
	^theKey! !

!ObjectIdTable methodsFor: 'copying'!

copy
	"Answer another instance just like the receiver."

	| id copy rf |
	id := indexDictionary.
	rf := referenceArray.
	referenceArray := referenceArray shallowCopy.
	indexDictionary := indexDictionary shallowCopy.
	copy := super shallowCopy.
	indexDictionary := id.
	referenceArray := rf.
	^copy! !

!ObjectIdTable methodsFor: 'private'!

findKeyOrNil: key  
	"Look for the key in the receiver.  If it is found, answer
	the index of the association containting the key, otherwise
	answer the index of the first unused slot."

	| index length probe pass |
	length := self basicSize.
	pass := 1.
	index := key identityHash \\ length + 1.
	[(probe := self basicAt: index) == nil or: [probe = key "Gids are compared by value"]] 
		whileFalse: [(index := index + 1) > length
				ifTrue: 
					[index := 1.
					pass := pass + 1.
					pass > 2 ifTrue: [^self grow findKeyOrNil: key]]].
	^index!

grow
	"Increase the number of elements of the collection."

	| newSelf |
	newSelf := self species new: self basicSize + self growSize.
	self tripleDo: 
		[:key :value :ref | newSelf noCheckAdd: key with: value with: ref ].
	self become: newSelf!

invariantViolated
	^self error: 'invariant violated'!

noCheckAdd: key with: value
	"Assume that the key is already in the receiver and store the value
	without checking to make sure."

	| index |
	index := self findKeyOrNil: key.
	self basicAt: index put: key.
	tally := tally + 1.
	valueArray at: index put: value.
	^indexDictionary at: value put: index!

noCheckAdd: key with: value with: ref
	"Assume that the key is already in the receiver and store the value
	without checking to make sure."

	| index |
	index := self findKeyOrNil: key.
	self basicAt: index put: key.
	tally := tally + 1.
	valueArray at: index put: value.
	referenceArray at: index put: ref.
	^indexDictionary at: value put: index!

rehash
	"Rebuild the receiver to ensure it is hashed correctly."
	"Smalltalk rehash."

	| newSelf |
	newSelf := self species new: self basicSize.
	self tripleDo: [:key :value :ref | newSelf noCheckAdd: key with: value with: ref ].
	self become: newSelf!

removeDangerouslyKey: key ifAbsent: errorBlock
	"Assume that the key is in the receiver and go fetch its value.
	If the value is nil, answer the result of evaluating the errorBlock.
	Otherwise, remove the key from the receiver."

	| location value oldKey length |
	location := self findKeyOrNil: key.
	(value := valueArray basicAt: location) == nil ifTrue: [^errorBlock value].
	indexDictionary removeKey: value ifAbsent: [self invariantViolated].
	self basicAt: location put: nil.
	valueArray at: location put: nil.
	referenceArray at: location put: 0.
	tally := tally - 1.
	length := self basicSize.
	[location := 
		location = length
			ifTrue: [1]
			ifFalse: [location + 1].
	(self basicAt: location) == nil]
		whileFalse: 
			[oldKey := self findKeyOrNil: (self basicAt: location).
			location = oldKey ifFalse: [self swap: location with: oldKey]]!

setTally
	super setTally.
	indexDictionary := IdentityDictionary new: self basicSize.
	referenceArray := ByteArray new: self basicSize.!

swap: oneLocation with: otherLocation 
	"Store otherLocation in the position for oneLocation, and
	vice versa."
	| oneValue otherValue oneIndex otherIndex |
	"find the indices"
	oneValue := self basicAt: oneLocation.
	otherValue := self basicAt: otherLocation.
	oneIndex := indexDictionary at: oneValue ifAbsent: [self invariantViolated].
	otherIndex := indexDictionary at: otherValue ifAbsent: [self invariantViolated].

	"consistency check"
	oneIndex ~= oneLocation ifTrue: [self invariantViolated].
	otherIndex ~= otherLocation ifTrue: [self invariantViolated].

	"swap the indices"
	indexDictionary at: oneValue put: otherIndex.
	indexDictionary at: otherValue put: oneIndex.

	"swap the reference count"
	referenceArray swap: oneLocation with: otherLocation.

	"swap the elements"
	^super swap: oneLocation with: otherLocation!

tripleDo: aBlock   

    1 to: self basicSize do: [ :index |
        (self basicAt: index) == nil
            ifFalse: [
                aBlock value: (self basicAt: index)
                       value: (valueArray at: index)
						value: (referenceArray at: index)
                ]
        ]! !

!ObjectIdTable methodsFor: 'dictionary removing'!

removeKey: key ifAbsent: errorBlock
	
	^self removeDangerouslyKey: key ifAbsent: [^errorBlock value]! !

!ObjectIdTable methodsFor: 'reference counting'!

decreaseRef: key 
	" Decreamental the remote reference count on the key
	if it is reaches 0, remove the object from the table"

	| index new |
	index := self findKeyOrNil: key.
	(self basicAt: index) isNil ifTrue: [^0].
	(new := (referenceArray at: index) - 2) = 0
		ifTrue: [self removeKey: key.]
		ifFalse: [referenceArray at: index put: (new := new bitAnd: 16rFF)].
	^new!

increaseRef: key 
	"Increase the remote reference count on the key.
	The remote references must not more than 128 
	 otherwise it won't work."

	| index old new |
	
	index := self findKeyOrNil: key.
	(self basicAt: index) isNil ifTrue: [^0].
	(old := referenceArray at: index) = 256
		ifTrue: [referenceArray at: index put: 1]
		ifFalse: [referenceArray at: index put: (new := old + 2 bitAnd: 255)].
	^new!

refNumber: key 
	"Reture the remote reference number on the key"
	| index |
	index := self findKeyOrNil: key.
	(self basicAt: index) isNil ifTrue: [^0].
	^referenceArray at: index! !

ReadWriteStream subclass: #PackingStream
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Remote-Invocation'!
PackingStream comment:
'Class BufferStream represents an accessor for a sequence of 
objects that communicate to the outside world.  
The instances of subclasses can contain non-homogenous elements.
Implementation of the class methods assumes streaming on a 
collection of binary, byte-sized elements;  
methods assume that a "word" consists of two-bytes.  
Its primary use is for prepare the communication data package.

		
		'!


!PackingStream methodsFor: 'accessing'!

collection
	^collection!

next: anInteger 
	"Answer the next anInteger elements of my collection.  Must override 
	because default uses self contents species, which might involve a Large 
	collection. "

	| newArray |
	newArray := collection species new: anInteger.
	1 to: anInteger do: [:index | newArray at: index put: self next].
	^newArray! !

!PackingStream methodsFor: 'nonhomogeneous positioning'!

padToNextWord
	"Make position even (on word boundary), answering the padding character if any."

	position even
		ifTrue: [^false]
		ifFalse: [^self next]!

padToNextWordPut: char
	"Make position even on word boundary, writing the padding character, 
char, if necessary.  Answer nil if already on a word boundary"

	position even
		ifTrue: [^nil]
		ifFalse: [^self nextPut: char]!

skipWords: nWords 
	"Position after nWords number of words."
	self skip: 2 * nWords!

wordPosition
	"Answer the current position in words."
	^self position / 2!

wordPosition: wp 
	"Set current position in words to be wp."
	self position: 2 * wp! !

!PackingStream methodsFor: 'nonhomogeneous accessing'!

nextFloat
	"Answer a single-precision (32 bit)
	floating point quantity"

	"The only way we can do this is by
	constructing the Float byte-by-byte."
	| f |
	f := 0.0 + 0.0.  "make a Float"
	f basicAt: 1 put: (self next).
	f basicAt: 2 put: (self next).
	f basicAt: 3 put: (self next).
	f basicAt: 4 put: (self next).
	^f!

nextFloat: aFloat
	"Store a single-precision (32 bit) 
	floating point quantity "

	"The only way we can do this is by 
	storing the Float byte-by-byte."
	| float |
	float := aFloat asFloat.
	self nextPut: (float basicAt: 1).
	self nextPut: (float basicAt: 2).
	self nextPut: (float basicAt: 3).
	self nextPut: (float basicAt: 4).
	^float!

nextGlobalId

	| gid size i |
	gid := GlobalId new.
	size := gid basicSize.
	i := 0.
	[(i := i + 1) <= size]
		whileTrue: [ gid basicAt: i  put: self next].
	^gid!

nextGlobalIdPut: gid

	| size i |
	size := gid basicSize.
	i := 0.
	[(i := i + 1) <= size]
		whileTrue: [ self nextPut: (gid basicAt: i)].!

nextIPSocketAddress

	| sa size i |
	sa := IPSocketAddress new.
	size := sa basicSize.
	i := 0.
	[(i := i + 1) <= size]
		whileTrue: [ sa byteAt: i  put: self next].
	^sa!

nextIPSocketAddressPut: sa

	| size i |
	size := sa basicSize.
	i := 0.
	[(i := i + 1) <= size]
		whileTrue: [ self nextPut: (sa byteAt: i)].!

nextLong
	"Answer a signed long (32 bit) quantity starting at the given byteIndex."

	| b0 b1 b2 w h |
	UninterpretedBytes isBigEndian
		ifTrue:
			[b0 := self next.  
			b1 := self next.  
			b2 := self next.   
			w := self next.  ]
		ifFalse:
			[b0 := self next.    
			b1 := self next.   
			b2 := self next.   
			w := self next.  ].
	"Following sequence minimizes LargeInteger arithmetic for small results."
	h := ((b0 bitAnd: 16r7F) - (b0 bitAnd: 16r80) bitShift: 8) + b1.
	b2=0 ifFalse: [w := (b2 bitShift: 8) + w].
	h=0 ifFalse: [w := (h bitShift: 16) + w].
	^w!

nextLongPut: aLong
	"Store a signed long (32 bit) quantity"
	| b0 b1 b2 b3 |
	b0 := aLong bitShift: -24.
	  b0 := (b0 bitAnd: 16r7F) - (b0 bitAnd: -16r80).
	b1 := (aLong bitShift: -16) bitAnd: 16rFF.
	b2 := (aLong bitShift: -8) bitAnd: 16rFF.
	b3 := aLong bitAnd: 16rFF.
	
	UninterpretedBytes isBigEndian
		ifTrue:
			["Only the first store can fail."
			self nextPut: b0.  
			self nextPut: b1.
			self nextPut: b2.  
			self nextPut: b3]
		ifFalse:
			["Only the first store can fail."
			self nextPut: b0.  
			self nextPut: b1.
			self nextPut: b2.  
			self nextPut: b3].
	^aLong!

nextNumber: n 
	"Answer the next n bytes as a positive Integer or LargePositiveInteger."

	"| s i |
	n <= 4 ifTrue: 
		[s := 0.
		i := 0.
		[(i := i + 1) <= n] whileTrue: [s := ((s bitShift: 8) bitOr: self next)].
		^s].
	s := LargePositiveInteger new: n.
	1 to: n do: [:j | s at: n + 1 - j put: self next].
	""reverse order of significance""
	^s truncated"

	| s |
	n <= 2
		ifTrue: 
			[s := 0.
			n timesRepeat: [s := s * 256 + self next].
			^s].
	s := Integer new: n neg: false.
	1 to: n do: [:i | s digitAt: n + 1 - i put: self next].
	"reverse order of significance"
	^s compressed!

nextNumber: n put: v 
	"Append to the receiver the argument, v, which is a positive SmallInteger or
	a LargePositiveInteger, as the next n bytes.  Possibly pad with leading zeros."

	| vlen i |
	n < (vlen := v digitLength) ifTrue: [self error: 'number too big'].

	"pad with leading zeros"
	i := n.
	[i > vlen] whileTrue: [self nextPut: 0. i := i - 1].
	i = 1 ifTrue: [^self nextPut: v].
	[i > 0] whileTrue: [self nextPut: (v digitAt: i). i := i - 1]!

nextSigned
	"Answer the next byte from the receiver as a signed byte."

	| value |
	self atEnd ifTrue: [^false].
	^(value := self next asInteger) > 127
		ifTrue: [256 + value negated]
		ifFalse: [value]!

nextSignedInteger
	"Answer the next two bytes from the receiver as a signed Integer."

	| returnValue |
	self atEnd ifTrue: [^false].
	returnValue := self next asInteger.
	self atEnd ifTrue: [^false].
	(returnValue := returnValue * 256 + self next asInteger) highBit = 16
		ifTrue:	[^returnValue - 65536	"16-bit negative 0"]
		ifFalse:	[^returnValue].!

nextString
	"Answer a new String read from the receiver. The first byte is the 
	length of the string, unless it is greater than 192, in which case the
	first two bytes encode the length. "

	| aString length i |
	length := self next.	"first byte."
	length >= 192 ifTrue: [length := length - 192 * 256 + self next].
	aString := String new: length.
	i := 0.
	[(i := i+1) <= length] 
		whileTrue: [aString basicAt: i put: self next].
	^aString!

nextStringPut: aString 
	"Append the string, aString, to the receiver."

	| length i |
	(length := aString size) < 192
		ifTrue: [self nextPut: length]
		ifFalse: 
			[self nextPut: (length // 256 + 192).
			self nextPut: (length \\ 256)].
	i := 0.
	[(i := i+1) <= length] 
		whileTrue: [ self nextPut: (aString basicAt: i)].
	^aString!

nextWord
	"Answer the next two bytes from the receiver as an Integer."

	| high low |
	high := self next.
		high==nil ifTrue: [^0].
	low := self next.
		low==nil ifTrue: [^0].
	^(high asInteger bitShift: 8) + low asInteger!

nextWordPut: aWord 
	"Append to the receiver an Integer as the next two bytes."

	self nextPut: ((aWord bitShift: -8) bitAnd: 255).
	self nextPut: (aWord bitAnd: 255).
	^aWord! !

!PackingStream methodsFor: 'positioning'!

resetContents
	"Reset pointers so that can restart reading the collection 
over which the receiver streams."
	position _ 0.
	readLimit _ 0! !

!PackingStream methodsFor: 'private'!

on: aCollection 

	collection := aCollection.
	readLimit := writeLimit := aCollection size.
	position := 0.
	self reset!

on: aCollection from: firstIndex to: lastIndex 

	| len |
	collection := aCollection.
	readLimit :=  lastIndex > (len := collection size)
						ifTrue: [len]
						ifFalse: [lastIndex].
	position := firstIndex <= 1
				ifTrue: [0]
				ifFalse: [firstIndex - 1]! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

PackingStream class
	instanceVariableNames: ''!


!PackingStream class methodsFor: 'instance creation'!

new
	^self basicNew! !

ByteArray variableByteSubclass: #GlobalId
	instanceVariableNames: ''
	classVariableNames: 'FixedLow ImmutableLow MaxImmutable MaxSpecial 
MaxUserSpecial MinImmutalbe MinSpecial MinUserSpecial NextId SpecialLow Step '
	poolDictionaries: ''
	category: 'Remote-Invocation'!
GlobalId comment:
'This class implements the unique global object identifier.
Each globalId is a 8 byte ByteArray and its value is a LargePositiveInteger

Immutable ids are used only for immutable objects and they have 
special high 4 byte value. The low 4 byte value of immutable ids 
must be supplied by immutable object creators (not generated by 
this class) because immutable objects may be replicated in many 
object contexts.

Special ids are used for special servicing or testing objects and they 
have special high 4 byte value. The low 4 byte value of special ids must be 
supplied by special object creators.
Special ids can be used by user to asign a well knwon or fixed value to 
a object gid.

Class Variables:
	NextId				<LargePositiveInteger> 		Next possible used globalId
	ImmutableHigh		<LargePositiveInteger>		High 4 byte value of  immutable ids
	SpecialHigh			<LargePositiveInteger>		High 4 byte value of special ids
 	Step					<SmallInteger>				Distance between the two nearest gids
	MaxImmutable 		<SmallInteger>				Maxmum immutalbe id value
	MaxSpecial 			<SmallInteger>				Maxmum special id value
	MaxUserSpecial 	<SmallInteger>				Maxmum user special id value
	MinImmutalbe 		<SmallInteger>				Minmum immutalbe id value
	MinSpecial 			<SmallInteger>				Minmum special id value
	MinUserSpecial		<SmallInteger>				Minmum user special id value
Note:
	GlobalIds are made difficault to guess because they may also be 
used as the capabilities (with message filters).
	Passing the gids across the object contexts in remote invcation, 
proxies will be created  automatically. For user, gids should be always 
encapsulated in proxies.

	There are arguments about the object globalId should contain 
object address.  Conceptually, object identity and object address are 
two defferent thing. Practically, combining the two has some advantages, 
e.g. to achieve the efficient mapping between object gid and address and 
to reduce the possible communication overhead. There are problems if 
combining the two, for example: should a object''s  gid be changed if the 
object migrate to another object context (its address changed)?
	Finally, I decide implemented the gid independently from the object 
address, which put things conceptually right and make the mapping between 
the gid and address (the object accessPoint) much more flexible.'!


!GlobalId methodsFor: 'private'!

highValue
	"get the high 4 bytes (5-8) value"

	^self doubleWordAt: 5!

highValue: aDoubleWord
	"set the high 4 bytes (5-8) value"

	^self doubleWordAt: 5 put: aDoubleWord!

identityHash
	"For dictionary look up"
	^self hash!

lowValue
	"get the low 4 bytes (1-4) value"

	^self doubleWordAt: 1!

lowValue: aDoubleWord
	"set the low 4 bytes (1-4) value"

	^self doubleWordAt: 1 put: aDoubleWord!

value
	"Return the bits in the receiver interpreted as an unsigned integer"

	| v i |
	v := LargePositiveInteger basicNew: 8.
	i := 0.
	[(i:= i + 1) <= 8 ] 
		whileTrue: [ v digitAt: i put: (self at: i)].
	^v truncated

	"GlobalId newId value"! !

!GlobalId methodsFor: 'testing'!

isImmutableId
	^(self lowValue = ImmutableLow)!

isSpecialId
	^(self lowValue = SpecialLow)! !

!GlobalId methodsFor: 'packing'!

putValueOn: aStream

	1 to: self basicSize do: [ :i | aStream nextPut: (self  basicAt: i)].! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

GlobalId class
	instanceVariableNames: ''!


!GlobalId class methodsFor: 'class initialization'!

initialize
	"GlobalId initialize"

	FixedLow := Time totalSeconds.
	NextId := 100.
	Step := 1.	
	ImmutableLow := 16r00000000.	
	SpecialLow :=16r00000000.

	"System special ids should from 0 to 999"
	"Known used system special ids:
	 RemoteInvocation NameServer 			101
	 RemoteInvocation AddressServer			100"

	MinSpecial := 0.
	MaxSpecial  := 999.

	"Immutable ids should from 1000 to 9,999"
	MinImmutalbe := 1000.
	MaxImmutable := 9999.

 	"User defined special ids should from 10,000 to 4,294,967,295 (16rFFFFFFFF)"
	MinUserSpecial := 10000.
	MaxUserSpecial := 16rFFFFFFFF.!

restart
	"reinitialize the NextId"

	FixedLow := Time totalSeconds.
	NextId := 100.! !

!GlobalId class methodsFor: 'instance creation'!

immutableId: aSmallInteger
	"Return a immutable object id"
	| nid  |
	aSmallInteger < MinImmutalbe | (aSmallInteger > MaxImmutable)
		ifTrue: [^self error: 'Non valid value for immutable ids'].
	nid :=  self basicNew: 8.
	nid doubleWordAt: 1 put: ImmutableLow; doubleWordAt: 5 put: aSmallInteger.
	^nid

"self immutableId: 5"!

new
	"New instance frame. Used by unpacking remote messages"

	^self basicNew: 8!

newId
	"GlobalId newId"

	| nid  |
	nid :=  self basicNew: 8.
	nid doubleWordAt: 1 put: FixedLow; doubleWordAt: 5 put: NextId.
	NextId := NextId + Step.
	^nid!

newValue: aNumber
	"New instance with a given value. 
	 Used at your own risk."

	| g i |
	g := self basicNew: 8.
	i := 0.
	[(i:= i + 1) <= aNumber size] 
		whileTrue: [ g digitAt: i put: (aNumber at: i)].
	^g!

specialId: aSmallInteger
	"Return a special object id"
	| nid  |
	aSmallInteger < MinSpecial | (aSmallInteger > MaxSpecial)
		ifTrue: [^self error: 'Non valid value for system special ids'].
	nid :=  self basicNew: 8.
	nid doubleWordAt: 1 put: SpecialLow; doubleWordAt: 5 put: aSmallInteger.
	^nid

"NGlobalId specialId: 5"!

userSpecialId: aSmallInteger
	"Return a special object id"
	| nid  |
	aSmallInteger < MinUserSpecial | (aSmallInteger > MaxUserSpecial)
		ifTrue: [^self error: 'Non valid value for user special ids'].
	nid :=  self basicNew: 8.
	nid doubleWordAt: 1 put: SpecialLow; doubleWordAt: 5 put: aSmallInteger.
	^nid

"GlobalId userSpecialId: 5"! !

!GlobalId class methodsFor: 'unpacking'!

getValueFrom: aStream

	| gid |
	gid := self new.
	1 to: gid basicSize do: [ :i | gid basicAt: i  put: aStream next].
	^gid! !

WeakDictionary variableSubclass: #WeakDic
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Remote-Invocation'!
WeakDic comment:
'I am the true WeakDICTIONARY not WeakIdentityDictionary that 
taken the wrong name of WeakDictionary'!


!WeakDic methodsFor: 'private'!

findKeyOrNil: key  
	"Look for the key in the receiver.  If it is found, answer
	the index of the association containing the key, otherwise
	answer the index of the first unused slot."

	| location length probe pass |
	length := self basicSize.
	pass := 1.
	location := key hash \\ length + 1.
	[(probe := self basicAt: location) == nil or: [probe = key]]
		whileFalse: 
			[(location := location + 1) > length
				ifTrue: 
					[location := 1.
					pass := pass + 1.
					pass > 2 ifTrue: [^self grow findKeyOrNil: key]]].
	^location! !

Object subclass: #RemoteInvocationServer
	instanceVariableNames: 'nameServer addressServer '
	classVariableNames: 'AddressDic BufferDic ClassCode DefaultServerName
 ExportObjectTable IamNameServer ImageName IncomeMessage MaxPackageSize 
MaxPackingCode MaxTag MessageProcessDispatchTable MetaCode MyAddress 
NameDic PackingCodes PassByGidCode PassByValueCode RecvProcess 
RemoteInvocationErrorSignal ReplyCodes SemDic Socket TimeoutSignal WellKnownPort '
	poolDictionaries: ''
	category: 'Remote-Invocation'!
RemoteInvocationServer comment:
'This class implements remote invocation server. It implements the 
forwarding-and-reply protocol used by remote invocation through the 
UDP/IP interface of the system. 

Because current Smalltalk is a implemented as single UNIX process, 
it is essencial to make the forwarding-and-reply unblocking while 
keeping the sychronized invocation interface towards the local message 
sender. It requires  a independent receiver thread from the sender thread, 
and remote invocation processing threads.  The sender thread will create a 
replySemaphore and wait at it after the message been forwared and the 
receiver (a independent thread) will signal the replySemanphore when it 
received the reply, the sender will wakeup and fetch the reply. 

When the receiver receives a remote invocation message, it will  fork a 
worker process (which is defined in the class method workOn) to actually 
handle the remote message, send it to a local object identified by the 
gid specified in the remote message and send back the invocation result.

When the remote inovcation start, it also set up several other servers: 
NameServer and  AddressServer either locally or remotely.

The NameServer maps  a name to exported object for the server and maps a 
name to a remote object (a proxy) for the client. It is basically for end 
user of distributed applications.

The AddressServer maps the object gid to object accessPoint (socketAddress). 
it is basically used by the remote invocation system itself.

This class has a solo instance RIServer.

Instance Variables: 

	addressServer 		<AddressDic/Proxy>	Mapping between gid and accessPoint
	nameServer 		<NameDic/Proxy>		Mapping between name and gid

Class Variables:

	AddressDic 			<Dictionary>		Mapping between gid and accessPoint
	BufferDic			<Dictionary>		Income message buffers.
	ClassCode			<SmallInteger>		Packing tag for classes.
	DefaultServerName <String>			The host name of NameServer
	DispatchTable		<Array>			Remote message processing method dispatch table.
	ExportObjectTable 	<Dictionary>		The Exported object table <gid, oop>
	ExportRefs			<IdentityDictionary> 	Keeps the remote reference number of exported objects. 
	IamNameServer	<Boolean>			This image is a name server or not.
	ImageName 		<String>			The name of this image
	IncomeMessage 	<RemoteMessage>	The income message buffer
	MaxPackageSize	<SmallInteger>		The maxmum udp package size - message header size (32)
	MaxPackingCode	<SmallInteger>		The largest packing code  can be.
	MaxTag				<SmallInteger>		Maxmum number of remote message types
	MessageProcessDispatchTable 	<Array>	Incoming message processing routine table.
	MetaCode			<SmallInteger>		The packing code for meta classes.
	MyAddress 			<SocketAddress>	My accessPoint
	NameDic			<Dictionary>		Mapping between name and gid
	PackingCodes 		<Array>			Packing/unpacking despatch table.
	PassByGidCode 	<SmallInteger>		Packing code for pass-by-gid objects
	PassByValueCode	<SmallInteger>		Packing code for pass-by-value objects
	RecvProcess 		<Process>			Receiver/Listener
	RemoteInvocationErrorSignal 	<Signal>	aSignal used to indicate that a remote invocation has a error reply
	ReplyCodes 		<Array>			Reply state codes for reply messages
	SemDic				<Dictionary>		Remote invocation message id dictionary
	Socket				<UnixSocketAccessor>	The UDP/IP socket
	TimeoutSignal		<Signal>			aSignal to indicate the remote invocation timeout error
	WellKnownPort 	<SmallInteger>		A port number used by NameServer image

Note:
	A server must be properly released before  it can be restarted 
which is request by external socket. You`v simplly got to tell UNIX 
that your are no longer using the socket because UNIX dose not do garbage 
collection.
	We still lack of the notion of DistributedObjectActivity which can 
be attached with all the activity dependent objects, e.g. control terminal, 
user id, etc..
	Distributed exception handling is still waiting to be added.

	Like all the other "distributed programs", sometimes, things can go mad!!!!!!!!!!!!!!!!!!!!!!!!
	Specially, when a proxy went into a deadloop of message forwarding. 
	Be carefull, you have been warned!!
'!


!RemoteInvocationServer methodsFor: 'object table accessing'!

exportObjByGid: gid ifAbsent: aBlock
	"Return the object by gid.
	If the requested object is not local, value the aBlock"

	
	^ExportObjectTable at: gid ifAbsent: [aBlock value]!

getObjectByName: aString
	"If the requested object is not local,
	return a proxy which represets a registered remote object "

	| gid add |
	gid := nameServer at: aString ifAbsent:[ ^nil ].
	gid == nil ifTrue:[^nil].
	^ ExportObjectTable at: gid ifAbsent:
			[add := addressServer at: gid.
			^Proxy newProxy: gid accessPoint: add]!

gidOf: obj
	"Return the gid of the object. 
	if the object has not got a gid yet, then asign a new gid to it.
	Sometimes, you have to make sure the object is not in the Globals table"

	| gid |
	^ExportObjectTable
		 keyAtValue: obj 
		ifAbsent: [gid := GlobalId newId.
					ExportObjectTable at: gid put: obj.
					^gid]! !

!RemoteInvocationServer methodsFor: 'register services'!

register: anObject name: aString
	"Register a local object with a name."

	| gid |
	self register: anObject name: aString globalId: (gid := GlobalId newId).
	^gid!

register: anObject name: aString globalId: gid
	"Register a local object with a name and a gId."
	
	nameServer removeKey: aString ifAbsent: [].
	nameServer at: aString put: gid.
	addressServer removeKey: gid ifAbsent: [].
	addressServer at: gid put: MyAddress.
	"All explicitly registered object are exported persistently"
	self exportPersistent: anObject withGid: gid.!

register: anObject name: aString userSpecialId: id
	"Register a local object with a name and a special Id."

	| gid |
	self registerService: anObject name: aString globalId: (gid := GlobalId userSpecialId: id).
	^gid!

registerServices
	"Regist local objects which supply services to other sites. 
	 When the remote invocation server starts, this method will 
	 be excuted automatically."

	self class registerServices!

removeByName: aString
	"remove the named service"

	| gid |
	gid := nameServer at: aString.
	nameServer removeKey: aString.
	addressServer removeKey: gid.
	ExportObjectTable removeKey: (ExportObjectTable keyAtValue: gid).!

syncNameAndAddress
	| newAddDic | 
	"Syncronize the AddressServer with the NameServer.
	Only used by the name server site."

	"RIServer syncNameAndAddress"

	
	newAddDic := Dictionary new: 100.
	nameServer do: [ :gid |
		newAddDic at: gid put: (addressServer at: gid)].
	addressServer become: newAddDic.! !

!RemoteInvocationServer methodsFor: 'exporting'!

export: anObject withGid: gid
	"export a local object with a gid."

	ExportObjectTable at: gid put: anObject.!

export: anObject withSpecialId: id 
	"export a local object with a special id number."

	| gid |

	gid := GlobalId specialId: id.
	(ExportObjectTable includesKey: gid)
		ifTrue: [ExportObjectTable removeKey: gid].
	ExportObjectTable at: gid put: anObject!

exportPersistent: anObject withGid: gid
	"export a local object with a gid."

	ExportObjectTable at: gid persistentPut: anObject.!

exportPersistent: anObject withSpecialId: id 
	"export a local object with a special id number."

	| gid |

	gid := GlobalId specialId: id.
	(ExportObjectTable includesKey: gid)
		ifTrue: [ExportObjectTable removeKey: gid].
	ExportObjectTable at: gid persistentPut: anObject! !

!RemoteInvocationServer methodsFor: 'distributed GC'!

sendDecRefOf: gid at: sa
	"Send a DECREF meta message to the 
	 accessPoint. No reply.
	 Message tag is 8."

	| remoteMessage |

	remoteMessage := RemoteMessage newheader.
	remoteMessage setGid: gid.
	remoteMessage setTag: 8.
	remoteMessage setSize: remoteMessage size.
	self forward: remoteMessage to: sa.
	^nil!

sendIncRefOf: gid at: sa timeout: t
	"Send a INCREF meta message to the
	accessPoint. 
	Message tag is 7."

	| delay replySemaphore timeout replySemaphoreId remoteMessage reply assoc |

	delay := Delay forSeconds: t.
	replySemaphore := Semaphore new.
	timeout := false.
	SemDic add: replySemaphore.
	replySemaphoreId := SemDic findElementOrNil: replySemaphore.
	remoteMessage := RemoteMessage newheader.
	remoteMessage setGid: gid.
	remoteMessage setTag: 7.
	remoteMessage setSize: remoteMessage size.
	remoteMessage setReplyId: replySemaphoreId.
	self forward: remoteMessage to: sa.
	
	[delay wait.
	timeout := true.
	replySemaphore signal]
		forkAt: Processor userInterruptPriority.
	replySemaphore wait.

	SemDic remove: replySemaphore ifAbsent:[ ].

	timeout ifTrue:[assoc := Association new.
					assoc key: 'ERROR'.
					assoc value: 'Remote invocation Timeout'.
					^assoc].
	"return the imcome message"
	reply := BufferDic at: replySemaphoreId ifAbsent: [].
	BufferDic removeKey: replySemaphoreId ifAbsent:[].
	assoc := self unpackReply: reply.
	"assoc key is reply tag, assoc value is the reply object"
	^assoc! !

!RemoteInvocationServer methodsFor: 'communication'!

forward: buf to: sa 
	"Send the the data in buf to the socket address sa"

	Object errorSignal
		handle: 
			[:ex |
			Transcript cr; show: ex errorString.
			self release.
			^nil]
		do: [Socket writeWait.
			Socket sendTo: sa buffer: buf]!

mforward: remoteMessage to: sa 
	"Forword the remote invocation message to the object  
	 identified by gid which is at the socket address sa 
	 If the message is too larger then sent it by multiple 
	 forwarding."


	| bufSize lastIndex header buf t1 t2 remainder index sec msg secSize tag |


	(remoteMessage getSize) <= MaxPackageSize
		ifTrue: [^self forward: remoteMessage to: sa]
		ifFalse: [
			lastIndex := 0.
			header := remoteMessage header.
			buf := remoteMessage body.
			bufSize := buf size.
			secSize := MaxPackageSize - remoteMessage headerSize.
			t1 := (tag := remoteMessage getTag) = 1
				ifTrue: [3]
				ifFalse: [tag = 2
					ifTrue:[5]
					ifFalse:[ tag = 10
						ifTrue:[11]
						ifFalse:[ RemoteInvocationErrorSignal raiseFrom: 'Invalid message tag']]].
			t2 := t1 + 1.
			[(remainder := bufSize - lastIndex) > 0]
				whileTrue: 
					[index := lastIndex + 1.
					sec := buf copyFrom: index 
								to: (lastIndex := remainder > secSize
											ifTrue: [lastIndex + secSize]
											ifFalse: [bufSize]).
					msg := header, sec.
					msg setTag: (lastIndex = bufSize 
										ifTrue:[t2]
										ifFalse:[t1]).
					msg setSize: msg size.
					self forward: msg to: sa]]!

receiver
	"I start a receiver thread.  
	 When the receiver receives a remote message,  
	 if the message is a forwarding message, a working process is 
	  forked to process remote invocation;  
	 if the message is a reply, the corespondent semaphore will  
	 be signaled and the processe waiting the reply will be wake 
	  up and  
	 get the reply message."

	| fa |

	fa := IPSocketAddress new.
	Transcript cr; show: 'RemoteInvocationServer starting'.
	[true]
		whileTrue: 
			["Wait for data arrival or timeout."
			Socket readWait.
			Socket receiveFrom: fa buffer: IncomeMessage.
			IncomeMessage setSenderAddress: fa.
			[self processIncomeMessage: IncomeMessage] fork.
			Processor yield].!

startReceiverAt: port 
	"Start a listenning server"

	Object errorSignal
		handle: 
			[:ex | 
			Transcript cr; show: ex errorString.
			self release.
			^nil]
		do: 
			[ 
			port == nil
				ifTrue: ["create a UDP socket without a preassigned port"
					Socket := UnixSocketAccessor newUDP]
				ifFalse: ["with a preassigned port"
					Socket := UnixSocketAccessor newUDPserverAtPort: port].
			MyAddress := Socket getName hostName: (UnixSocketAccessor getHostname).
			SemDic := IdentitySet new: 100.
			BufferDic := Dictionary new: 100.
			IncomeMessage := RemoteMessage newAsBuffer.
			RecvProcess := [self receiver]
						forkAt: Processor userInterruptPriority.]! !

!RemoteInvocationServer methodsFor: 'packing-unpacking'!

getClassFrom: stream
	"Get the local class by name"
	
	| classname class classformat code |
	code := stream next.
	classname := stream nextString asSymbol.
	classformat := stream nextLong.
	class := Smalltalk at: classname ifAbsent:[self inconsistentPacking].
	class isBehavior 
		ifFalse:[self inconsistentPacking].
	class instancesSentByValue & (classformat = class format)
		ifFalse:[self inconsistentPacking].
	^code = MetaCode
			ifTrue:[class class]
			ifFalse:[class]!

inconsistentPacking

self halt.
	self error: 'Inconsistent packing protocol'!

packCopyReply: replyCode reply: reply
	"Packing up the object structure recursively 
	 For remote copy"

	| buf stream  |
	stream := PackingStream on: (buf := RemoteMessage new: 200).
	stream position: (buf messageStart) -1 .
	stream nextPut: replyCode.
	Packer deepPack: reply on: stream.
	^stream contents!

packMessage: message 
	"Pack the message in a byteArray"

	| stream selector arguments numberOfArguments arg buf |
	message isNil ifTrue: [^RemoteMessage newheader].
	stream := PackingStream on: (buf := RemoteMessage new: 200).
	stream position: buf messageStart - 1.
	selector := message selector.
	arguments := message arguments.
	numberOfArguments := arguments size.
	stream nextStringPut: selector asString.
	stream nextWordPut: numberOfArguments.
	(1 to: numberOfArguments)
		do: [:index | (arg := arguments at: index) isSentByValue
				ifTrue: 
					[self putClassCodeOf: arg on: stream.
					arg putValueOn: stream]
				ifFalse: 
					[stream nextPut: PassByGidCode.
					stream nextGlobalIdPut: (self gidOf: arg).
					stream nextIPSocketAddressPut: MyAddress]].
	^stream contents!

packReply: replyCode reply: obj 
	"pack a reply in a byteArray"

	| stream buf   |
	stream := PackingStream on: (buf := RemoteMessage new: 200).
	stream position: buf messageStart - 1.
	stream nextPut: replyCode.
	obj isSentByValue
		ifTrue: 
			[self putClassCodeOf: obj on: stream.
			obj putValueOn: stream]
		ifFalse: 
			[stream nextPut: PassByGidCode.
			stream nextGlobalIdPut: (self gidOf: obj).
			stream nextIPSocketAddressPut: MyAddress].
	^stream contents!

putClassCodeOf: obj on: stream 
	"Put the class code or name of the object class on the aStream. 
	The obj must be SentByValue object."

	| classCode class |
	classCode := PackingCodes indexOf: obj class ifAbsent: [PassByValueCode].
	stream nextPut: classCode.
	classCode = PassByValueCode
		ifTrue: 
			[class := obj class.
			class isMeta
				ifTrue: 
					[stream nextPut: MetaCode.
					stream nextStringPut: class soleInstance name]
				ifFalse: 
					[stream nextPut: ClassCode.
					stream nextStringPut: class name].
			stream nextLongPut: class format]!

unpackCopyReply: buffer
	"Packing up the object structure recursively  
	 For remote copy"

	| stream reply  |

	stream := PackingStream
				on: buffer
				from: buffer messageStart
				to: buffer size.
	reply := Association new.
	reply key: (ReplyCodes at: stream next).
	reply value: (Unpacker nextObjectOn: stream).
	^reply!

unpackMessage: buffer 
	"Unpack a message in a byteArray"

	| stream selector arguments arg numberOfArgs class code |
	stream := PackingStream
				on: buffer
				from: buffer messageStart
				to: buffer size.
	selector := stream nextString asSymbol.
	numberOfArgs := stream nextWord.
	arguments := Array new: numberOfArgs.
	(1 to: numberOfArgs)
		do: 
			[:index | 
			(code := stream next) == PassByValueCode
				ifTrue: 
					[class := self getClassFrom: stream.
					arg := class getValueFrom: stream]
				ifFalse: [arg := (PackingCodes at: code)
								getValueFrom: stream].
			arguments at: index put: arg].
	^Message selector: selector arguments: arguments!

unpackReply: buffer 
	"Unpack a reply in a byteArray"

	| stream reply class code |
	stream := PackingStream
				on: buffer
				from: buffer messageStart
				to: buffer size.
	reply := Association new.
	reply key: (ReplyCodes at: stream next).
	(code := stream next) == PassByValueCode
		ifTrue: 
			[class := self getClassFrom: stream.
			reply value: (class getValueFrom: stream)]
		ifFalse: [reply value: ((PackingCodes at: code)
					getValueFrom: stream)].
	^reply! !

!RemoteInvocationServer methodsFor: 'accessing'!

accessPoint
	"Return my socket name"

	^MyAddress!

imageName

	^ImageName!

objectIdType
	^18 		"ProxyType"!

setImageName: aString
	"RemoteInvocationServer setImageName: 'test'"

	^ImageName := aString! !

!RemoteInvocationServer methodsFor: 'initialize-release'!

initialize
	"Start local remote invocation server. 
	If the image name you chosed is the host name then
	this image will be start as a NameServer otherwise
	this image will  NOT  be a NameServer.
	If this image is not a NameServer, it must using two 
	remote objects as the NameServer and the AddressServer.
	Set up the name server, address server, className server  etc. "
	"RemoteInvocationServer start"

	"Start server"
	| hostName serverName serverAddr    |

	IamNameServer := false.
	hostName := ImageName.

	"Initialize the ProxyTable. (On ever run)?"
	Proxy initializeProxyTable.
	
	"Reinitialize the GlobalId start value"
	GlobalId restart.

	"Initialize the ExportObjectTable"
	ExportObjectTable := ObjectIdTable new: 256.

	"Setup the name for this image"
	ImageName == nil
		ifTrue: [ImageName := hostName].
	Transcript cr; show: 'The name of this image is ', ImageName.

	serverName :=
		 DialogView 
					request: 'Where is the NAME SERVER 
(enter the host name)?' 
			initialAnswer: (DefaultServerName isNil
					ifTrue: ['sake']
					ifFalse: [DefaultServerName]).
	serverName isEmpty ifTrue: [^self].
	DefaultServerName := serverName.

	Transcript cr; show: 'Name server is ', serverName.
	
	"Setup the remote invocation listenServer"
	self startReceiverAt: nil.

	"Setup the servers"
	serverAddr := IPSocketAddress 
					hostName: serverName
					port: WellKnownPort.
	nameServer := Proxy newProxy: (GlobalId specialId: 100) 
								accessPoint: serverAddr.
	addressServer := Proxy newProxy: (GlobalId specialId: 101) 
								accessPoint: serverAddr.!

initializeAsNameServer
	"Start local remote invocation server. 
	This image will be used as a NameServer.
	This image name will be same as host name.
	Currently there is only one NameServer allowed in a LAN and 
	This NameServer must be keep running to allow other images 
	to communicated each other.

	Set up the name server, address server, className server 
	locally "
	"RemoteInvocationServer start"

	"Start server"
	| hostName     |

	IamNameServer := true.
	hostName := ImageName.

	"Initialize the ProxyTable. (On ever run)?"
	Proxy initializeProxyTable.
	
	"Reinitialize the GlobalId start value"
	GlobalId restart.

	"Initialize the ExportObjectTable"
	ExportObjectTable := ObjectIdTable new: 256.


	"Setup the name for this image"
	ImageName := hostName.
	DefaultServerName := hostName.

	"Setup the remote invocation listenServer"
	self startReceiverAt: WellKnownPort.

	"Setup the servers"
	nameServer := NameDic.
	addressServer := AddressDic.
	self exportPersistent: nameServer  withSpecialId: 100.
	self exportPersistent: addressServer withSpecialId: 101.
	"self registerServices."
	Transcript cr; show: 'The name of this image is ', hostName; 
				cr; show: 'This is a NameServer image.'.!

release
	"RemoteInvocationServer release"
	Proxy releaseProxyTable.
	RecvProcess class == Process ifTrue:[ RecvProcess terminate].
	RecvProcess := nil.
	Socket ~~ nil ifTrue: [
		Socket close.
		Socket := nil].
	Transcript cr; show: 'RemoteInovcationServer shut down'.! !

!RemoteInvocationServer methodsFor: 'remote invocation'!

forwardAndGetReply: aMessage to: gid at: sa timeout: t 
	"Forword the remote invocation message to the object identified by 
	gid which is at the socket address sa and wait for the reply. "
	| replySemaphoreId replySemaphore timeout delay reply remoteMessage assoc  |

	delay := Delay forSeconds: t.
	replySemaphore := Semaphore new.
	timeout := false.
	SemDic add: replySemaphore.
	replySemaphoreId := SemDic findElementOrNil: replySemaphore.
	remoteMessage := self packMessage: aMessage.
	remoteMessage setGid: gid.
	remoteMessage setTag: 1.
	remoteMessage setSize: remoteMessage size.
	remoteMessage setReplyId: replySemaphoreId.	
	self mforward: remoteMessage to: sa.
	[delay wait.
	timeout := true.
	replySemaphore signal] forkAt: Processor userInterruptPriority.
	replySemaphore wait.

	SemDic remove: replySemaphore ifAbsent:[ ].

	timeout ifTrue:[^TimeoutSignal raiseFrom: t].

	"return the imcome message"
	reply := BufferDic at: replySemaphoreId ifAbsent:[].
	BufferDic removeKey: replySemaphoreId.
	assoc := self unpackReply: reply.
	assoc key = 'ERROR' 
		ifTrue:[^RemoteInvocationErrorSignal raiseFrom: assoc value.].
	"assoc key is reply tag, assoc value is the reply object"
	^assoc!

forwardMessage: aMessage to: gid at: sa
	"Forword the remote invocation message to the object identified by 
	gid which is at the socket address sa, Do not need a reply. "
	| remoteMessage |

	remoteMessage := self packMessage: aMessage.
	remoteMessage setGid: gid.
	remoteMessage setTag: 10.
	remoteMessage setSize: remoteMessage size.
	self mforward: remoteMessage to: sa.
	^nil!

remoteCopy: gid at: sa timeout: t 
	" Get a deepcopy of the remote object identified by gid."

	| replySemaphoreId replySemaphore timeout delay reply remoteMessage assoc |

	delay := Delay forSeconds: t.
	replySemaphore := Semaphore new.
	timeout := false.
	SemDic add: replySemaphore.
	replySemaphoreId := SemDic findElementOrNil: replySemaphore.
	remoteMessage := self packMessage: nil.
	remoteMessage setGid: gid.
	remoteMessage setTag: 9.
	remoteMessage setSize: remoteMessage size.
	remoteMessage setReplyId: replySemaphoreId.	

	"[self mforward: remoteMessage to: sa] fork."
	self mforward: remoteMessage to: sa.

	[delay wait.
	timeout := true.
	replySemaphore signal] fork.
	replySemaphore wait.

	SemDic remove: replySemaphore ifAbsent:[ ].

	timeout ifTrue:[^TimeoutSignal raiseFrom: t].

	"return the imcome message"
	reply := BufferDic at: replySemaphoreId ifAbsent:[].
	BufferDic removeKey: replySemaphoreId.
	assoc := self unpackCopyReply: reply.
	assoc key = 'ERROR' 
		ifTrue:[^RemoteInvocationErrorSignal raiseFrom: assoc value.].
	"assoc key is reply tag, assoc value is the reply object"
	^assoc!

testforwardAndGetReply: aMessage to: gid at: sa timeout: t 
	"Forword the remote invocation message to the object identified by 
	gid which is at the socket address sa and wait for the reply. "
	| replySemaphoreId replySemaphore timeout delay reply remoteMessage assoc time |
time := Time millisecondsToRun: [
	delay := Delay forSeconds: t.
	replySemaphore := Semaphore new.
	timeout := false.
	SemDic add: replySemaphore.
	replySemaphoreId := SemDic findElementOrNil: replySemaphore.
	remoteMessage := self packMessage: aMessage.
	remoteMessage setGid: gid.
	remoteMessage setTag: 1.
	remoteMessage setSize: remoteMessage size.
	remoteMessage setReplyId: replySemaphoreId.	].
Transcript cr; show: ' Packing Timing: ' , (time )  printString , ' milliseconds.'.
"Transcript cr; show: aMessage printString."
time := Time millisecondsToRun: [
	self mforward: remoteMessage to: sa.
	[delay wait.
	timeout := true.
	replySemaphore signal] forkAt: Processor userInterruptPriority.
	replySemaphore wait.].
Transcript cr; show: ' Forwarding Timing: ' , (time )  printString , ' milliseconds.'.

time := Time millisecondsToRun: [
	SemDic remove: replySemaphore ifAbsent:[ ].

	timeout ifTrue:[^TimeoutSignal raiseFrom: t].

	"return the imcome message"
	reply := BufferDic at: replySemaphoreId ifAbsent:[].
	BufferDic removeKey: replySemaphoreId.
	assoc := self unpackReply: reply.].
Transcript cr; show: ' Unpack Timing: ' , (time )  printString , ' milliseconds.'; cr; cr.
	assoc key = 'ERROR' 
		ifTrue:[^RemoteInvocationErrorSignal raiseFrom: assoc value.].
	"assoc key is reply tag, assoc value is the reply object"
	^assoc! !

!RemoteInvocationServer methodsFor: 'message processing'!

decreaseRef: msg 
	"Decrease the remote reference count of
	a exported object.  No reply."


	| gid |
	gid := msg getGid.
	"Transcript cr; show: 'RefDEC for ', gid printString."
	ExportObjectTable decreaseRef: gid.
	^nil!

increaseRef: msg 
	"Increase the remote reference count of
	a exported object. "

	| reply gid |
	gid := msg getGid.
	"Transcript cr; show: 'RefINC for ' , gid printString."
	reply := ExportObjectTable increaseRef: gid.
	reply := self packReply: (ReplyCodes indexOf: 'SUCCESS')
				reply: reply.
	reply setTag: 2.
	reply setSize: reply size.
	reply setReplyId: msg getReplyId.
	self mforward: reply to: msg getSenderAddress!

lastToSender: msg 
	"Income message is the last segmentation to a waiting thread"

	| sid  sem old |

	sid := msg getReplyId.
	sem := SemDic basicAt: sid.
	sem ~~ nil
		ifTrue: [
			old := BufferDic at: sid ifAbsent: [RemoteMessage new].
			BufferDic at: sid put: old , msg body.
			SemDic remove: sem.
			sem signal]!

lastToServer: msg 
	"Income message is the last segmentation to server"

	| sid old  |

	sid := msg getSenderAddress.
	sid ~~ nil
		ifTrue: [
			old := BufferDic at: sid ifAbsent: [RemoteMessage new].
			BufferDic removeKey: sid.
			self workOn: old , msg body]!

lastToServerNoReply: msg 
	"Income message is the last segmentation to server"

	| sid old  |

	sid := msg getSenderAddress.
	sid ~~ nil
		ifTrue: [
			old := BufferDic at: sid ifAbsent: [RemoteMessage new].
			BufferDic removeKey: sid.
			self workOnWithoutReply: old , msg body]!

processIncomeMessage: msg 
	"Process the ccording to the incomeMessage tag"

	| tag |

	tag := msg getTag.
	tag <= MaxTag
		ifTrue:[
			^self perform: (MessageProcessDispatchTable at: tag) with: (msg copyFrom: 1 to: msg getSize)].
	^self error: 'Remote message type error'!

segmentToSender: msg 
	"Income message is a segmentation to a waiting thread"

	| sid old sem |

	sid := msg getReplyId.
	sem := SemDic basicAt: sid.
	sem ~~ nil
		ifTrue: [
			old := BufferDic at: sid ifAbsent: [RemoteMessage new].
			old size = 0
				ifTrue: [BufferDic at: sid put: msg]
				ifFalse: [BufferDic at: sid put: old , msg body]]!

segmentToServeNoReplyr: msg 
	"Income message is a segmentation to server"

	^self segmentToServer: msg!

segmentToServer: msg 
	"Income message is a segmentation to server"

	| sid old |

	sid := msg getSenderAddress.
	sid ~~ nil
		ifTrue: [
			old := BufferDic at: sid ifAbsent: [RemoteMessage new].
			old size = 0
				ifTrue: [BufferDic at: sid put: msg]
				ifFalse: [BufferDic at: sid put: old , msg body]]!

sendObject: msg 
	"Sending a deepcopy of the object to the remote site"

	| gid obj reply |

	
	gid := msg getGid.
	obj := ExportObjectTable at: gid ifAbsent: [].
	obj == nil
		ifTrue: [
			reply := self packReply: (ReplyCodes indexOf: 'ERROR' )  reply: 'Nonexist remote object.'.
			Transcript cr; show: 'Non-public gid: ' , gid printString; cr]
		ifFalse: [
			reply := self packCopyReply: (ReplyCodes indexOf: 'SUCCESS' )  reply: obj].
	reply setTag: 2.
	reply setSize: reply size.
	reply setReplyId: msg getReplyId.
	self mforward: reply to: msg getSenderAddress!

toSender: msg 
	"Income message is address to a waiting thread"

	| sid sem |

	sid := msg getReplyId.
	sem := SemDic basicAt: sid.
	sem ~~ nil
		ifTrue: [
			BufferDic at: sid put: msg.
			sem signal]!

workOn: remoteMessage 
	"This is the remote message processing thread body.  
	 Will be forked by the communicator when a remote  
	 invocation happens"

	| message gid obj reply   |

	gid := remoteMessage getGid.
	obj := ExportObjectTable at: gid ifAbsent: [].
	obj isReallyProxy
		ifTrue: [reply := self packReply:(ReplyCodes indexOf:  'APCHANGED') reply: obj]
		ifFalse: [obj == nil
				ifTrue: [
					reply := self packReply: (ReplyCodes indexOf: 'ERROR' )  reply: 'Nonexist remote object.'.
					Transcript cr; show: 'Non-public gid: ' , gid printString; cr]
				ifFalse: [
					Object errorSignal
						handle: 
							[:ex |
							reply := self packReply: (ReplyCodes indexOf: 'ERROR' )   reply: ex errorString.
							reply setReplyId: remoteMessage getReplyId.
							reply setTag: 2.
							reply setSize: reply size.
							self mforward: reply to: remoteMessage getSenderAddress.
							^nil]
						do: 
							[message := self unpackMessage: remoteMessage.
							reply := obj perform: message selector withArguments: message arguments].
					reply := self packReply: (ReplyCodes indexOf: 'SUCCESS') reply: reply]].
	reply setTag: 2.
	reply setSize: reply size.
	reply setReplyId: remoteMessage getReplyId.
	self mforward: reply to: remoteMessage getSenderAddress!

workOnWithoutReply: remoteMessage 
	"This is the remote message processing thread body. 
	Will be forked by the communicator when a remote 
	invocation happens. No reply is sent back."

	| message gid obj |
	gid := remoteMessage getGid.
	obj := ExportObjectTable at: gid ifAbsent: [].
	obj isReallyProxy | (obj == nil)
		ifTrue: [^nil]
		ifFalse: [Object errorSignal handle: [:ex | ^nil]
				do: 
					[message := self unpackMessage: remoteMessage.
					obj perform: message selector withArguments: message arguments]].
	^nil! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

RemoteInvocationServer class
	instanceVariableNames: ''!


!RemoteInvocationServer class methodsFor: 'class initialization'!

initialize
	"RemoteInvocationServer initialize"

	"Each running Smalltalk-80 image should has its own name "
	ImageName := UnixSocketAccessor getHostname.
	
	Smalltalk at: #RIServer put: nil.
	NameDic := Dictionary new: 256.
	AddressDic := Dictionary new: 256.
	WellKnownPort := 2005.
	MaxPackageSize := 4000.

	TimeoutSignal := (Object errorSignal newSignalMayProceed: true)
		notifierString: 'Remote Invocation Timeout: ';
		nameClass: self message: #timeoutSignal.

	RemoteInvocationErrorSignal := (Object errorSignal newSignalMayProceed: true)
		notifierString: 'Remote Invocation Error: ';
		nameClass: self message: #remoteInvocationSignal.

	"Remote message processing dispatch table"
	MessageProcessDispatchTable := #(
							workOn:						"tag = 1"
							toSender:						"tag = 2"
							segmentToServer:				"tag = 3"
							lastToServer:					"tag = 4"
							segmentToSender:				"tag = 5"
							lastToSender:					"tag = 6"
							increaseRef:					"tag = 7"
							decreaseRef:					"tag = 8"
							sendObject:					"tag = 9"
							workOnWithoutReply:			"tag = 10"
							segmentToServerNoReply:		"tag = 11"
							lastToServerNoReply:			"tag = 12"
							).
	MaxTag := 12.

	"PackingCodes for primitive immutable classes"
	PackingCodes := Array new: 25.
	PackingCodes at: 1 put: Object.
	PackingCodes at: 2 put: Character.
	PackingCodes at: 3 put: Float.
	PackingCodes at: 4 put: GlobalId.
	PackingCodes at: 5 put: IntegerArray.
	PackingCodes at: 6 put: IPSocketAddress.
	PackingCodes at: 7 put: LargeNegativeInteger.
	PackingCodes at: 8 put: LargePositiveInteger.
	PackingCodes at: 9 put: SmallInteger.
	PackingCodes at: 10 put: String.
	PackingCodes at: 11 put: UndefinedObject.
	PackingCodes at: 12 put: UninterpretedBytes.
	PackingCodes at: 13 put: Proxy.
	PackingCodes at: 14 put: True.
	PackingCodes at: 15 put: False.
	PackingCodes at: 16 put: Double.
	PackingCodes at: 17 put: Symbol.
	PackingCodes at: 18 put: ByteEncodedString.
	PackingCodes at: 19 put: ByteString.
	PackingCodes at: 20 put: ISO8859L1String.
	PackingCodes at: 21 put: MacString.
	PackingCodes at: 22 put: ByteSymbol.
	PackingCodes at: 23 put: TwoByteSymbol.
	PackingCodes at: 24 put: TwoByteString.
	MaxPackingCode := 24.
	PassByValueCode := 0.
	PassByGidCode := PackingCodes indexOf: Proxy.

	ReplyCodes := #(
				'SUCCESS'
				'ERROR'
				'APCHANGED'
				).

	MetaCode := 1.
	ClassCode := 2.! !

!RemoteInvocationServer class methodsFor: 'fileOut'!

fileOut
	"RemoteInvocationServer fileOut"

	| file selectors fname |
	fname := DialogView request: 'file out remove invocation stuff on which filename?' 
							initialAnswer: 'RemoteInvocation.v5.st'.
	fname isEmpty ifTrue: [^self].
	file := (Filename named: fname) writeStream.
	file timeStamp; 
		cr; cr; 
		nextPutAll: '| s | s := '; 
		store: self title; 
		nextPutAll: '.
	Transcript show: s.!!'; 
		cr; cr.	
	"Clean out previous versions"
	file nextPutAll: '
	(DialogView confirm:
''It is recommended that the previous version of this
package be removed.  Doing so involves removing all
methods in category ''''remote invocation packing unpacking'''' 
and removing classes in Remote-Invocation and all their subclasses.
Do you wish to do so now?'') ifTrue: 
			[#(''remote invocation'' ''packing'' ''unpacking'') do:
					[:category | category asSymbol.
					Smalltalk allBehaviorsDo: 
						[:class | (class organization listAtCategoryNamed: category) do:
							[:sel | class removeSelector: sel]]].
			(Smalltalk at: #Proxy ifAbsent:[]) isNil
				ifFalse:[Proxy releaseProxyTable].
			#(''Remote-Invocation'' ) 
				do: [:cat | (SystemOrganization superclassOrder:  cat )do: 
						[:class | Smalltalk removeClassNamed: class name]]]!!'.

	file cr; cr.
	"File out required support methods (explicit change set)"
	OrderedCollection new
		"add: (IdentityDictionary -> #at:ifAbsentPutValue:);"
		do: [:assoc|
			assoc key
				fileOutMessage: assoc value
				on: file
				moveSource: false
				toFile: nil.
			Smalltalk changes removeSelectorChanges: assoc value class: assoc key].

	"prerequested classes for remote invocation"
	#( ) do:
		[:class | (Smalltalk at: class) fileOutOn: file].
	"remote invocation classes"
	#('Remote-Invocation' ) do: 
		[:cate | SystemOrganization fileOutCategory: cate asSymbol on: file].
	#(#'remote invocation' #packing #unpacking ) do: 
		[:cat | Smalltalk classNames do: 
			[:cn | | class |
			((class := Smalltalk at: cn) ~~ self and: [(class inheritsFrom: self) not])
				ifTrue: [(Array with: class with: class class)
						do: 
							[:fileOutClass | 
							selectors := fileOutClass organization listAtCategoryNamed: cat.
							selectors size > 0
								ifTrue: 
									[Transcript cr; show: fileOutClass name.
									fileOutClass
										fileOutCategory: cat
										on: file
										moveSource: false
										toFile: nil.
									(Smalltalk changes respondsTo: #removeClassChanges:method:)
										ifTrue: [selectors do: [:selector | Smalltalk changes removeClassChanges: class method: selector]]]]]]].
	file close!

printOut
	"RemoteInvocationServer printOut"

	| textStream|
	textStream := TextStream on: (String new: 1024).

	textStream cr; cr; emphasis: 5; nextPutAll: self title; emphasis: 1; nextPut: Character newPage.

	#('Distribution' ) do: [:cate | 	SystemOrganization printOutCategory: cate on: textStream].

	textStream nextPut: Character newPage.
	#(#'remote invocation' ) do: [:category  | 
	Smalltalk classNames do: [:cn|
		| class |
		((class := Smalltalk at: cn) ~~ self
		and: [(class inheritsFrom: self) not]) ifTrue: [

			(Array with: class with: class class) do: [:fileOutClass|
				(fileOutClass organization listAtCategoryNamed: category) size > 0 ifTrue: [
					fileOutClass printOutCategory: category on: textStream]]]]].

	Document new
		setDefaultHeader;
		startParagraph;
		addText: textStream contents asText;
		close;
		toPrinter!

title
	^'
		Remote Invocation Version 1.5
		For ObjectWorks/Smalltalk-80R4

		Wang Xu.
		Queen Mary and Westfield College.
		University of London.
		14 June 1992.
'! !

!RemoteInvocationServer class methodsFor: 'register services'!

registerServices
	"Regist local objects which  supply services to other sites.
	User or appliction program can put thier service objects 
	with a chosed name in the Services dictionary and when a 
	remote invocation server started, all the objects in teh Services
	will be registered under the given name."
 
	|  | 

	RIServer isNil ifTrue:[^nil].

	"Remote games"
	"RIServer register: NetworkGameServer new name: (ImageName, 'NetworkGameServer')."

	"For remote browsing"
	RIServer register: SystemOrganization name: (ImageName,'Organization' ).

	"For testing"
	RIServer register: 1 name: (ImageName, 'Test1').
	RIServer register: 2 name: (ImageName, 'Test2').
	RIServer register: TestObject new name: (ImageName, 'Test4').
	RIServer register: Transcript name: (ImageName, 'Transcript').
	
	"For testing shared HotDraw Editor"
	"drawing := PersistentDrawing newOn: 'bar'.
	RIServer register: drawing name: 'bar' globalId: drawing gid. "

	IamNameServer ifTrue: [RIServer syncNameAndAddress].! !

!RemoteInvocationServer class methodsFor: 'initialize-release'!

release
	"RemoteInvocationServer release"

	RIServer release.
	RIServer := nil.!

start
	"Start local remote invocation server. 
	If the image name you chosed is the host name then
	this image will be start as a NameServer otherwise
	this image will  NOT  be a NameServer.
	If this image is not a NameServer, it must using two 
	remote objects as the NameServer and the AddressServer.
	Set up the name server, address server, className server  etc. "
	"RemoteInvocationServer start"
	
	RIServer isNil 
		ifFalse: [ RIServer release ].
	RIServer := self new.
	RIServer  initialize!

startAsNameServer
	"Start local remote invocation server. 
	This image will be used as a NameServer.
	This image name will be same as host name.
	Currently there is only one NameServer allowed in a LAN and 
	This NameServer must be keep running to allow other images 
	to communicated each other."
	"RemoteInvocationServer startAsNameServer"

	RIServer isNil 
		ifFalse: [ RIServer release ].
	RIServer := self new.
	RIServer initializeAsNameServer! !

!RemoteInvocationServer class methodsFor: 'example'!

clientExample1
	"RemoteInvocationServer clientExample1"

	| a   complexPara  time i j testImageName |

	testImageName := 'sake'.

	a := RIServer getObjectByName: (testImageName,'Test4').

	complexPara := #('FirstElement' 'SecondElement' ).
	time :=0.
	j := 50.
	i := j.	
	[(i := i - 1) > 0]
		whileTrue: 
			[time := time + (Time millisecondsToRun: [a testMessage1: complexPara])].
	Transcript cr; show: ' Average time of each call: ' , (time/j) truncated printString , ' milliseconds.'!

clientExample1a
	|   aProxy   complexPara reply time i testImageName |
	
	testImageName := 'sake'.

	aProxy := RIServer getObjectByName: (testImageName,'Test4').

	complexPara := Array new: 2.
	complexPara at: 1 put: 'FirstElement'.
	complexPara at: 2 put: 'SecondElement'.
	i := 11.
	[(i := i - 1) > 0]
		whileTrue: 
			[time := Time millisecondsToRun: [reply := aProxy testMessage1: complexPara.].
			Transcript cr; show: reply printString.
			Transcript cr; show: ' Timing: ' , time printString , ' milliseconds.'].
	
"RemoteInvocationServer clientExample1a"!

clientExample1b
	|   aProxy   complexPara reply testImageName |
	
	testImageName := 'sake'.

	aProxy := RIServer getObjectByName: (testImageName,'Test4').

	complexPara := Array new: 2.
	complexPara at: 1 put: 'FirstElement'.
	complexPara at: 2 put: 'SecondElement'.
	reply := aProxy testMessage1: complexPara.
	Transcript cr; show: (reply printString).

"RemoteInvocationServer clientExample1b"!

clientExample2
	|    aProxy bProxy     testImageName |
	
	testImageName := 'sake'.

	aProxy := RIServer getObjectByName: (testImageName,'Test1').
	bProxy := RIServer getObjectByName: (testImageName,'Test2').

	Transcript cr; show: (aProxy * 123456789) printString.
	Transcript cr; show: (bProxy * 123456789) printString.

"RemoteInvocationServer clientExample2"!

clientExample3
	|    aProxy bProxy     testImageName |
	
	testImageName := 'sake'.

	aProxy := RIServer getObjectByName: (testImageName,'Test1').
	bProxy := RIServer getObjectByName: (testImageName,'Test2').


	[true]
		whileTrue: 
			[Transcript cr; show: (aProxy * 123456789) printString.
			Transcript cr; show: (bProxy * 123456789) printString.].

"RemoteInvocationServer clientExample3"!

clientExample4
	"RemoteInvocationServer clientExample4 "
	| aProxy testImageName i reply time  j |
	
	testImageName := 'sake'.

	aProxy := RIServer getObjectByName: (testImageName,'Test1').

	i := 50.
	j := i.
	time := Time millisecondsToRun: [
		[(i := i - 1) > 0]
			whileTrue: 
				[reply := aProxy + 1]].
	Transcript cr; show: ' Average Timing: ' , (time // j)  printString , ' milliseconds.'.! !

!RemoteInvocationServer class methodsFor: 'documentation'!

distributedGarbageCollection
	"
	We have a simple and lightweight distributed garbage collection algorithm ready to work.
	It uses remote reference counting.

		In each image, all the remote references to the same remote object is represented by
		a single proxy. All the proxies will be keeped in aWeekDictionary.

		Remote invocation server in each image maintains a remote reference count for each 		exported object that indictes how may proxies the object has (e.g. how may images 
		currently using the object). 

		In the initialization of a proxy, a IncreaseRemoteReference message must be 
		successfully returned otherwise the proxy (representing a remote pointer) 
		creation will be fail and nil is returned.
		
		When there is no local reference to a proxy, the proxy will be finalized by the 		weekDictionary. The finalize message of a proxy will decreamental the remote 
		reference count.

		If the remote reference count of any exported object decreased to zero, the object
		will removed from the exported object table and from then it is up to the local 
		garbage collector to decide wether it is should be reclaimed.

	Currently the remote garbage is turned off.
	Because each time the remote invocation start, it will refresh the exported object table and
	the proxy table. In fact, all the remote garbages (or non-garbages) are gone by the restart the
	remote invocation server.

	If you do feel  the distributed garbage collection is a necessary then turn it on by
	seting the Proxy class variable DistributedGarbageCollectionOn to true (in the Proxy
	class method initialize).
	
	Note: 
		This distributed garbage collection algorithm will fail to reclaim 
		the distributed circular garbage ante the garbages with a outstanding 
		remote reference count due to the network failure. 
		It only guarantee that a non-garbage remote object will never be reclaimed by
		the distrubuted garbage collector unless the remote invocation server restart. 
	"!

errorProcessing
	"
	A remote invocation may raise a timeout signal if the timeout condition is reached. 
	The remote invocation timeout value is set by proxies.
	User can customize it by:
			aProxy proxySetTimeout: seconds

	The remote invocation timeout signal can be catched and handled in the 
	following way:

		RemoteInvocationServer timeoutSignal
			handle: [ :ex |
					timeout exception processing]
			do: [
				normal processing ]
	
	Should a remote invocation cause a error no the remote site, the error message will be
	passed back and a RemoteInvoationErrorSignal is raised with the error message.
	The RemoteInvocationErrorSignal can be catched and handled in the following way:

		RemoteInvocationServer remoteInvocationErrorSignal
			handle: [ :ex |
					remote invocation error exception processing ]
			do: [
				normal processing ]
	
	"
	^self shouldNotImplement!

essentialConcept

	"Read the comment of this class"
	
	^self shouldNotImplement!

installation
	"This method explains necessary steps 
	 to install the remote invocation into a  
	 exsiting image after fileIn. 
	
	Prerequested classes for remote invocation
				IPSocketAddress 
				SocketAddress 
				UDSocketAddress 
				UnixSocketAccessor

	If you have not installed these classes then fileIn: 
			/official/smalltalk/ObjectworksR4/utils/UnixSock.st

	 1. Setup the image name 

	 	Each distributed Smalltalk image must has a name <String>. 
		The default image name is the hostname of the unix system.
		If your image is not a NameServer, you may setup the image 
		name explicitly  in the RemoteInvocationServer class method 
		initialize: 
			 ImageName := 'yourImageName'. 
	 	Then do RemoteInvocationServer initialize. 
	  
	 	If the image will be the NameServer, then the image name will 
		be automatically reset by the system, and the image name will be 
		the same as the hostname.  
	  
	2. 	In a distributed Smalltalk environment, there can be many
		interracted images each running a remote invocation server
		to forward and receive the remote messages.
		One and only one image must act as a NameServer, which 
		resposible for mapping names and addresses of distributed objects.
		Before other distributed image can communicated with each
		other, the NameServer must be running.
		To start a remote invocation server as name server, do:
	
				RemoteInvocationServer startAsNameServer

		To start a common invocation server, do: 

				RemoteInvocationServer start

		it will prompt and ask you the name of the host on which
		the NameServer is running.
		To terminate a remote invocation server, do: 

				RemoteInvocationServer release
		
		After a remote server is running it is named by the global
		variable RIServer.
		
	3. 	Register Services
		An object can be exported and registered as a public service,
		which can be shared by any remote users.
		You can register an object by:

				RIServer register: anObject name: aString.

		The anObject is publicated object and the aString is the name
		to remember it.	 The name must be unique through out all the 
		Distritributed Smalltalk images otherwise the old object pulicated
		under the same name will be replaces by the new one.

		You may put a set of predefined public objects and their names
		in the RemoteInvocationServer class methods: 
				registerServices.
		Everytime you do:
				RemoteInvocationServer registerServices.
		All the predefined public objects in the method will be
		registered.
		
		Every time you reboot a distributed Smalltalk image, all the 
		public objects have to be registed again because they may have
		new accessPoints due to the socket address is changed.

		You may find adding the following sub menus in LauncherView menu may 
		be helpfull: (add to LauncherView class method: initializeLauncherMenu)
			
			keys := #('Start server' 'Register services' 'Kill the server' 'Start As NameServer').
			subMenu := PopUpMenu
				labelList: (Array with: keys)
				values: (Array new: keys size).
			subMenu fontKey: defaultFontKey.
			subMenu valueAt: 1 put: [RemoteInvocationServer start].
			subMenu valueAt: 2 put: [RemoteInvocationServer registerServices].
			subMenu valueAt: 3 put: [RemoteInvocationServer release].
			subMenu valueAt: 4 put: [RemoteInvocationServer startAsNameServer].
			LauncherMenu valueAt: <a proper index> put: subMenu.

			keys := #('remote browser' 'remote inspector' ).
			subMenu := PopUpMenu
				labelList: (Array with: keys)
				values: (Array new: keys size).
			subMenu fontKey: defaultFontKey.
			subMenu valueAt: 1 put: [RemoteBrowser remoteBrowser].
			subMenu valueAt: 2 put: [RemoteInspector remoteInspect].
			LauncherMenu valueAt: <a proper index> put: subMenu.

			
	 "

	
	^self shouldNotImplement "Documentation only"!

securityProblems
	"
	There is a big security hole on the remote invocation.
	User must be aware that any exporeted object may accept any message
	it can understand, for example:

		aProxy class perform: nastyThing

	Currently the NameDictionary and AddressDictionary and all other testing
	and demonstration exported objects are not secured.

	All exported object really should be carefully desinded and protected (for example,
	encapsulated by message a filter). 

	How to protect a larger number of shared objects still remains as a research topic.

	Still, this package should not be used in a security sensitive envirement.
	
	YOU HAVE BEEN WARNED!!

	"!

sentByValueObjects
	"This method explains how an object is passed
	as parameters in a remote invocation
	 
	By default, all the arguments in a remote message are passed
	by gids and at the destination and proxies will be automatically
	created by remote invocation server.

	A set of premitive objects are defined as isSentByValue,
	e.g. ArithmeticValue, ByteArray, String, etc., 
	which means those objects  will be passed by their values in 
	remote invocation messages. 

	In some case, a user may define a class of object  isSentByValue 
	for speeding up the remote invocation (or slow it down ;-)
	or for other operational reasons. The user can put the class method
		instancesSentByVaule
				^true.
	in the class and let it returns true. User must also define two methods:
			putValueOn: aBufferStream for packing the object value and
			getValueFrom: aBufferStream for reconstructing the object.
	putValueOn: must be an instance message and getValueFrom: must be
	a class message. The remote invocation server will automatically 
	using these two methods packing and unpacking user defined sent-by-value
	objects and passes these objects by value.

	The binary format of isSentByValue objects can be defined by user and
	implememted by putValueOn:/getValueFrom:. However, the binary format
	must be consistent between the putValueOn: and getValueFrom: in order to
	pack and unpack object correctly.
	
	Generally, objects passed by value have consistency problem, 
	that is the two copies on both the souce and destination site 
	may become defferent in the future. The remote invocation server 
	does not guarantee the consistency of isSentByValue copies. 
	 "

	^self shouldNotImplement!

usingRemoteObject
	"This method explains how to registed a
	public object in the server image and
	use it in a client image
	  
	
	1. 	Register an object in the server site:
		An object can be exported and registered as a public service,
		which can be shared by any remote users.
		You can register an object by:

				RIServer register: anObject name: aString.

		RIServer is a global variable that points to a instance of RemoteInvocationServer.
		The anObject is publicated object and the aString is the name
		to remember it.	 The name must be unique through out all the 
		Distritributed Smalltalk images otherwise the old object pulicated
		under the same name will be replaces by the new one.
	
		Every time you reboot a distributed Smalltalk image, the objects have 
		to be registed again because it may have new accessPoints due to the 
		communication socket address is changed.

	2. 	Using an remote object in the client site: 
		An proxy of the remote object can be obtained by:

				obj := RIServer getObjectByname: aString.

		The obj can then be used in the same way as a local object:

				result := obj aMessage.

		See the examples in the protocol example of the 		RemoteInvocationServer class.
	 "


	^self shouldNotImplement! !

!RemoteInvocationServer class methodsFor: 'Signal constants'!

remoteInvocationErrorSignal
	"Answer the Signal used to indicate that
	a remote invocation has a error reply"


	^RemoteInvocationErrorSignal!

timeoutSignal
	"Answer the Signal used to indicate that
	a remote invocation is timeout"

	^TimeoutSignal! !
RemoteMessage initialize!
Proxy initialize!
RemoteInvocationServer initialize!
Encapsulator initialize!
GlobalId initialize!


!Boolean class methodsFor: 'remote invocation'!

instancesSentByValue
	"When instances of the receiver are arguments in a remote message
	 answer whether they should be sent by value or by proxy.
	 Instances of the receiver should be well known and immutable
	 so should be sent by value." 
	^true! !

!Character class methodsFor: 'remote invocation'!

instancesSentByValue
	"When instances of the receiver are arguments in a remote message
	 answer whether they should be sent by value or by proxy.
	 Instances of the receiver should be well known and immutable
	 so should be sent by value." 
	^true! !

!ClassDescription methodsFor: 'remote invocation'!

instancesSentByValue
	"By default, object are not sent by value in the remote invocation
	but sent by global id.
	Subclasses should overwrite this method if their instances are
	needed pass by value and user must define the packing/unpacking
	methods, e.g. putValueOn: and getValueFrom:"

	^false! !

!Integer class methodsFor: 'remote invocation'!

instancesSentByValue
	"When instances of the receiver are arguments in a remote message
	 answer whether they should be sent by value or by proxy.
	 Instances of the receiver should be well known and immutable
	 so should be sent by value." 
	^true! !

!IntegerArray class methodsFor: 'remote invocation'!

instancesSentByValue

	^true! !

!LimitedPrecisionReal class methodsFor: 'remote invocation'!

instancesSentByValue
	"When instances of the receiver are arguments in a remote message
	 answer whether they should be sent by value or by proxy.
	 Instances of the receiver should be well known and immutable
	 so should be sent by value." 
	^true! !

!Object methodsFor: 'remote invocation'!

isReallyProxy
	^false!

isSentByValue
	^self class instancesSentByValue!

remoteClass
	^self class! !

!Object class methodsFor: 'remote invocation'!

instancesSentByValue
	"By default, object are not sent by value in the remote invocation
	but sent by global id.
	Subclasses should overwrite this method if their instances are
	needed pass by value and user must define the packing/unpacking
	methods, e.g. putValueOn: and getValueFrom:"

	^false! !

!Proxy methodsFor: 'remote invocation'!

isSentByValue
	"When the receiver is a argument in a remote message 
	 answer whether they should be sent by value or by proxy. 
	 The receiver should be well known and immutable 
	 so should be sent by value."


	^true!

putValueOn: stream
	stream nextGlobalIdPut: globalId.
	stream nextIPSocketAddressPut: accessPoint.! !

!String class methodsFor: 'remote invocation'!

instancesSentByValue
	"When instances of the receiver are arguments in a remote message
	 answer whether they should be sent by value or by proxy.
	 Instances of the receiver should be well known and immutable
	 so should be sent by value." 
	^true! !

!UndefinedObject class methodsFor: 'remote invocation'!

instancesSentByValue
	"When instances of the receiver are arguments in a remote message
	 answer whether they should be sent by value or by proxy.
	 Instances of the receiver should be well known and immutable
	 so should be sent by value." 
	^true! !

!UninterpretedBytes class methodsFor: 'remote invocation'!

instancesSentByValue
	
	^true! !

!Boolean methodsFor: 'packing'!

putValueOn: aStream
	"do nothing"! !

!Character methodsFor: 'packing'!

putValueOn: stream
	"Characters are stored as a tag byte followed by their ascii value in 2 bytes."

	stream
		nextPut: (self asciiValue bitShift: -8);
		nextPut: (self asciiValue bitAnd: 16rFF)! !

!Double methodsFor: 'packing'!

putValueTo: stream
	stream nextNumber: 8 put: self! !

!Float methodsFor: 'packing'!

putValueOn: stream 
	"Always store floats in bigendian order"

	UninterpretedBytes isBigEndian
		ifTrue: [stream nextPut: (self basicAt: 1); nextPut: (self basicAt: 2);nextPut: (self basicAt: 3); nextPut: (self basicAt: 4)]
		ifFalse: [stream nextPut: (self basicAt: 4); nextPut: (self basicAt:3);nextPut: (self basicAt: 2); nextPut: (self basicAt: 1)]! !

!GlobalId methodsFor: 'packing'!

putValueOn: aStream

	1 to: self basicSize do: [ :i | aStream nextPut: (self  basicAt: i)].! !

!IntegerArray methodsFor: 'packing'!

putValueOn: aStream

	aStream nextStringPut: self asString! !

!IPSocketAddress methodsFor: 'packing'!

putValueOn: aStream

	aStream nextIPSocketAddressPut: self! !

!LargeNegativeInteger methodsFor: 'packing'!

putValueOn: stream 
	| i basicSize |
	i _ 0.
	stream nextNumber: 4 put: (basicSize _ self basicSize).
	[(i _ i + 1) <= basicSize]
		whileTrue: [stream nextPut: (self basicAt: i)]! !

!LargePositiveInteger methodsFor: 'packing'!

putValueOn: stream 
	| i basicSize |
	i _ 0.
	stream nextNumber: 4 put: (basicSize _ self basicSize).
	[(i _ i + 1) <= basicSize]
		whileTrue: [stream nextPut: (self basicAt: i)]! !

!PositionableStream methodsFor: 'packing'!

deepPackObject: obj
	"Pack the object on myself."

	obj  deepPackOn: self!

deepUnpackObject
	^Object new unpackQueue: nil from: self!

unpackObject
	"Object ::= ClassId ObjectBody 
	ObjectBody has references to other objects except primitives" 

	^Object new unpackFrom: self! !

!SmallInteger methodsFor: 'packing'!

putValueOn: stream 
	"SmallIntegers are stored as their value with the 32nd bit set as a 
	tag."


	stream
		nextPut: (((self bitShift: -24) bitAnd: 16rFF) bitOr: 16r80);
		nextPut: ((self bitShift: -16) bitAnd: 16rFF);
		nextPut: ((self bitShift: -8) bitAnd: 16rFF);
		nextPut: (self bitAnd: 16rFF)! !

!String methodsFor: 'packing'!

putValueOn: aStream
	aStream nextWordPut: self size.
	1 to: self size do: [:i | aStream nextPut: (self at: i) asCharacter ]! !

!UndefinedObject methodsFor: 'packing'!

putValueOn: stream
	^self! !

!UninterpretedBytes methodsFor: 'packing'!

putValueOn: stream 
	| size |
	size _ self basicSize.
	stream nextNumber: 4 put: size.
	(1 to: size)
		do: [:i | stream nextPut: (self byteAt: i)]! !

!Character class methodsFor: 'unpacking'!

getValueFrom: stream
	^self value: (stream next bitShift: 8) + stream next! !

!Double class methodsFor: 'unpacking'!

getValueFrom: stream

	"The only way we can do this is by
	constructing the Double byte-by-byte."
	| f |
	f := 0 asDouble.  "make a Double"
	1 to: f basicSize do:
		[ :i |	f basicAt: i put: (stream next)].
	^f! !

!False class methodsFor: 'unpacking'!

getValueFrom: stream
	^false! !

!Float class methodsFor: 'unpacking'!

getValueFrom: stream 
	| float   |
	float _ self basicNew: 4.
	UninterpretedBytes isBigEndian
		ifTrue: [float basicAt: 1 put: stream next; basicAt: 2 put: stream next; basicAt: 3 put: stream next; basicAt: 4 put: stream next]
		ifFalse: [float basicAt: 4 put: stream next; basicAt: 3 put: stream next; basicAt: 2 put: stream next; basicAt: 1 put: stream next].
	^float! !

!GlobalId class methodsFor: 'unpacking'!

getValueFrom: aStream

	| gid |
	gid := self new.
	1 to: gid basicSize do: [ :i | gid basicAt: i  put: aStream next].
	^gid! !

!IntegerArray class methodsFor: 'unpacking'!

getValueForm: aStream
	^self fromString: aStream nextString! !

!IPSocketAddress class methodsFor: 'unpacking'!

getValueFrom: aStream

	^aStream nextIPSocketAddress! !

!LargeNegativeInteger class methodsFor: 'unpacking'!

getValueFrom: stream 
	| i basicSize value |
	i _ 0.
	basicSize _ stream nextNumber: 4.
	value _ self basicNew: basicSize.
	[(i _ i + 1) <= basicSize]
		whileTrue: [value basicAt: i put: stream next].
	^value! !

!LargePositiveInteger class methodsFor: 'unpacking'!

getValueFrom: stream 
	| i basicSize value |
	i _ 0.
	basicSize _ stream nextNumber: 4.
	value _ self basicNew: basicSize.
	[(i _ i + 1) <= basicSize]
		whileTrue: [value basicAt: i put: stream next].
	^value! !

!SmallInteger class methodsFor: 'unpacking'!

getValueFrom: stream 
			| value |
	(value _ (stream next bitAnd: 16r7F)) > 16r3F
		ifTrue: [value _ value - 16r80].
	value _ (value bitShift: 8) bitOr: stream next.
	value _ (value bitShift: 8) bitOr: stream next.
	value _ (value bitShift: 8) bitOr: stream next.
	^value! !

!String class methodsFor: 'unpacking'!

getValueFrom: aStream
	| newString length |
	newString := self new: (length := aStream nextWord).
	1 to: length do: [:i | newString at: i put: aStream next asCharacter].
	^newString! !

!Symbol class methodsFor: 'unpacking'!

getValueFrom: stream
	^(String getValueFrom: stream) asSymbol! !

!True class methodsFor: 'unpacking'!

getValueFrom: stream
	^true! !

!UndefinedObject class methodsFor: 'unpacking'!

getValueFrom: stream
	^nil! !

!UninterpretedBytes class methodsFor: 'unpacking'!

getValueFrom: stream 

	|  size buf |
	size _ stream nextNumber: 4.
	buf _ ByteArray new: size.
	(1 to: size)
		do: [:i | buf at: i put: stream next].
	^self from: buf! !
